import itertools

import numpy as np
import pandas as pd

from sdmxthon.model.dataset import Dataset
from sdmxthon.utils.handlers import add_list
from sdmxthon.utils.parsing_words import SERIES, OBS, ID, STRSPE, GENERIC, \
    SERIESKEY, ATTRIBUTES, VALUE, OBS_DIM, OBSVALUE, OBSKEY, DIM_OBS, \
    exc_attributes

chunksize = 50000


def get_element_to_list(data, mode):
    obs = {}
    if VALUE in data[mode]:
        data[mode][VALUE] = add_list(data[mode][VALUE])
        for k in data[mode][VALUE]:
            obs[k[ID]] = k[VALUE.lower()]
    return obs


def process_df(test_list: list, df: pd.DataFrame):
    if len(test_list) > 0:
        if df is not None:
            df = pd.concat([df, pd.DataFrame(test_list)], ignore_index=True)
        else:
            df = pd.DataFrame(test_list)

    del test_list[:]

    return test_list, df


def reading_generic_series(dataset) -> pd.DataFrame:
    # Generic Series
    test_list = []
    df = None
    dataset[SERIES] = add_list(dataset[SERIES])
    for series in dataset[SERIES]:
        keys = dict()
        # Series Keys
        if not isinstance(series[SERIESKEY][VALUE], list):
            series[SERIESKEY][VALUE] = [series[SERIESKEY][VALUE]]
        for v in series[SERIESKEY][VALUE]:
            keys[v[ID]] = v[VALUE.lower()]
        if ATTRIBUTES in series:
            if not isinstance(series[ATTRIBUTES][VALUE], list):
                series[ATTRIBUTES][VALUE] = [series[ATTRIBUTES][VALUE]]
            for v in series[ATTRIBUTES][VALUE]:
                keys[v[ID]] = v[VALUE.lower()]
        if not isinstance(series[OBS], list):
            series[OBS] = [series[OBS]]
        for data in series[OBS]:
            obs = dict()
            obs[OBS_DIM] = data[OBS_DIM][VALUE.lower()]
            obs[OBSVALUE.upper()] = data[OBSVALUE][VALUE.lower()]
            if ATTRIBUTES in data:
                obs = {**obs, **get_element_to_list(data, mode=ATTRIBUTES)}
            test_list.append({**keys, **obs})
        if len(test_list) > chunksize:
            test_list, df = process_df(test_list, df)

    test_list, df = process_df(test_list, df)

    return df


def reading_generic_all(dataset) -> pd.DataFrame:
    # Generic All Dimensions
    test_list = []
    df = None
    dataset[OBS] = add_list(dataset[OBS])
    for data in dataset[OBS]:
        obs = dict()
        obs = {**obs, **get_element_to_list(data, mode=OBSKEY)}
        if ID in data[OBSVALUE]:
            obs[data[OBSVALUE][ID]] = data[OBSVALUE][VALUE.lower()]
        else:
            obs[OBSVALUE.upper()] = data[OBSVALUE][VALUE.lower()]
        obs = {**obs, **get_element_to_list(data, mode=ATTRIBUTES)}
        test_list.append({**obs})
        if len(test_list) > chunksize:
            test_list, df = process_df(test_list, df)

    test_list, df = process_df(test_list, df)

    return df


def reading_str_series(dataset) -> pd.DataFrame:
    # Structure Specific Series
    test_list = []
    df = None
    dataset[SERIES] = add_list(dataset[SERIES])
    for data in dataset[SERIES]:
        keys = dict(itertools.islice(data.items(), len(data) - 1))
        if not isinstance(data[OBS], list):
            data[OBS] = [data[OBS]]
        for j in data[OBS]:
            test_list.append({**keys, **j})
        if len(test_list) > chunksize:
            test_list, df = process_df(test_list, df)

    test_list, df = process_df(test_list, df)

    return df


def get_at_att_str(dataset):
    return {k: dataset[k] for k in dataset if k not in exc_attributes}


def get_at_att_gen(dataset):
    attached_attributes = {}
    if VALUE in dataset[ATTRIBUTES]:
        dataset[ATTRIBUTES][VALUE] = add_list(dataset[ATTRIBUTES][VALUE])
        for k in dataset[ATTRIBUTES][VALUE]:
            attached_attributes[k[ID]] = k[VALUE.lower()]
    return attached_attributes


def create_dataset(dataset, metadata, global_mode):
    if STRSPE == global_mode:
        # Dataset info
        attached_attributes = get_at_att_str(dataset)

        # Parsing data
        if SERIES in dataset:
            # Structure Specific Series
            df = reading_str_series(dataset)
        elif OBS in dataset:
            # Structure Specific All dimensions
            df = pd.DataFrame(dataset[OBS]).replace(np.nan, '')
        else:
            df = pd.DataFrame()
    elif GENERIC == global_mode:

        # Dataset info
        if ATTRIBUTES in dataset:
            attached_attributes = get_at_att_gen(dataset)
        else:
            attached_attributes = {}

        # Parsing data
        if SERIES in dataset:
            # Generic Series
            df = reading_generic_series(dataset)
            renames = {'OBSVALUE': 'OBS_VALUE',
                       'ObsDimension': metadata[DIM_OBS]}
            df.rename(columns=renames, inplace=True)
        elif OBS in dataset:
            # Generic All Dimensions
            df = reading_generic_all(dataset)
            df.replace(np.nan, '', inplace=True)
            df.rename(columns={'OBSVALUE': 'OBS_VALUE'}, inplace=True)
        else:
            df = pd.DataFrame()
    else:
        raise Exception
    dataset = Dataset(attached_attributes=attached_attributes, data=df)

    return dataset



"""
    DataSet file contains only the class DataSet, used for external assets
    to perform any validation or conversion to different formats.
"""

import json
from datetime import date, datetime

import pandas as pd
from pandas import DataFrame

from sdmxthon.model.definitions import DataStructureDefinition, \
    DataFlowDefinition
from sdmxthon.model.header import Header
from sdmxthon.parsers.data_validations import validate_data
from sdmxthon.parsers.write import writer
from sdmxthon.utils.enums import MessageTypeEnum


class Dataset:
    """ An organised collection of data.

    :param structure: Associates the DataStructureDefinition to the DataSet
    :type structure: class:`DataStructureDefinition`

    :param dataflow: Associates the DataFlowDefinition to the Dataset
    :type dataflow: class:`DataFlowDefinition`

    :param dataset_attributes: Contains all the attributes from the DataSet \
    class of the Information Model
    :type dataset_attributes: dict

    :param attached_attributes:  Contains all the attributes at a Dataset level
    :type attached_attributes: dict

    :param data: Any object compatible with pandas.DataFrame()
    :type data: `Pandas Dataframe \
    <https://pandas.pydata.org/pandas-docs/stable \
    /reference/api/pandas.DataFrame.html>`_
    """

    def __init__(self, structure: DataStructureDefinition = None,
                 dataflow: DataFlowDefinition = None,
                 dataset_attributes: dict = None,
                 attached_attributes: dict = None, data=None):

        self._dataset_attributes = {}

        self.data = None
        self._dataflow = None
        self._structure = None
        self.attached_attributes = {}

        if attached_attributes is not None:
            self.attached_attributes = attached_attributes.copy()

        if structure is not None and dataflow is not None:
            raise ValueError('A Dataset cannot have a structure '
                             'and a dataflow, use only one')
        if structure is not None:
            self._dataflow = None
            self.structure = structure
        elif dataflow is not None:
            self.dataflow = dataflow

        if data is not None:
            if isinstance(data, pd.DataFrame):
                self.data = data.copy()
            else:
                self.data = pd.DataFrame(data)

        if dataset_attributes is None:
            self._check_DA_keys({})
        else:
            self._check_DA_keys(dataset_attributes)

    def __str__(self):
        if self.structure is not None:
            return f'<DataSet  - {self.structure.id}>'

        return '<DataSet - No Structure found>'

    def __unicode__(self):
        if self.structure is not None:
            return f'<DataSet  - {self.structure.id}>'

        return '<DataSet - No Structure found>'

    def __repr__(self):
        if self.structure is not None:
            return f'<DataSet  - {self.structure.id}>'

        return '<DataSet - No Structure found>'

    @property
    def dataflow(self) -> DataFlowDefinition:
        """Associates the DataFlowDefinition to the Dataset

        :class: `DataFlowDefinition`

        """
        return self._dataflow

    @dataflow.setter
    def dataflow(self, value: DataFlowDefinition):
        if isinstance(value, DataFlowDefinition) or value is None:
            self._dataflow = value
            if value is not None:
                if len(self.attached_attributes) > 0:
                    attached_attributes = self.attached_attributes.copy()
                    for k in attached_attributes.keys():
                        if k not in value.structure.dataset_attribute_codes:
                            del self._attached_attributes[k]

                    for k in value.structure.dataset_attribute_codes:
                        if k not in attached_attributes.keys():
                            raise ValueError(f'Missing attribute {k} '
                                             f'at a dataset level '
                                             f'(attached_attributes)')

                if ('OBS_VALUE' in self._data.columns and
                        value.structure.measure_code != 'OBS_VALUE'):
                    self._data = self._data.rename(
                        {'OBS_VALUE': value.structure.measure_code})
                self.dataset_attributes['setId'] = value.id
                self._structure = value
                self._structure = value.structure
        else:
            raise TypeError('dataflow must be a DataFlowDefinition')

    @property
    def structure(self) -> DataStructureDefinition:
        """Associates the DataStructureDefinition to the DataSet

        :class: `DataStructureDefinition`

        """
        if self._structure is None and self.dataflow is not None:
            return self.dataflow.structure
        return self._structure

    @structure.setter
    def structure(self, value: DataStructureDefinition):
        if not isinstance(value,
                          DataStructureDefinition) and value is not None:
            raise TypeError('structure must be a DataStructureDefinition')

        if self._dataflow is not None and value is not None:
            raise ValueError('dataflow property is not None')

        if value is not None:
            if len(self.attached_attributes) > 0:
                attached_attributes = self.attached_attributes.copy()
                for k in attached_attributes.keys():
                    if k not in value.dataset_attribute_codes:
                        del self._attached_attributes[k]

                for k in value.dataset_attribute_codes:
                    if k not in attached_attributes.keys():
                        raise ValueError(f'Missing attribute {k} at a '
                                         f'dataset level '
                                         f'(attached_attributes)')

            if ('OBS_VALUE' in self._data.columns and
                    value.measure_code != 'OBS_VALUE'):
                self._data = self._data.rename(
                    {'OBS_VALUE': value.measure_code})
            self.dataset_attributes['setId'] = value.id
        self._structure = value

    @property
    def dataset_attributes(self) -> dict:
        """Contains all the attributes from the DataSet class of the
        `Information Model <https://sdmx.org/wp-content/uploads/SDMX_2-1
        -1_SECTION_2_InformationModel_201108.pdf#page=85>`_

        :class: dict

        """
        return self._dataset_attributes

    @dataset_attributes.setter
    def dataset_attributes(self, value: dict):
        self._check_DA_keys(value)

    @property
    def attached_attributes(self) -> dict:
        """Contains all the attributes at a Dataset level with
        NoSpecifiedRelationship """
        return self._attached_attributes

    @attached_attributes.setter
    def attached_attributes(self, value: dict):
        if self.structure is not None:
            temp = value.copy()
            for k in temp.keys():
                if k not in self.structure.dataset_attribute_codes:
                    raise ValueError(f'{k} not in the attributes at dataset '
                                     f'level for DSD '
                                     f'{self.structure.unique_id}')

        self._attached_attributes = value

    @property
    def data(self) -> pd.DataFrame:
        """Pandas DataFrame that withholds all the data"""
        return self._data

    @data.setter
    def data(self, value):
        if value is None:
            self._data = pd.DataFrame()
        else:
            if isinstance(value, pd.DataFrame):
                temp = value.copy()
            else:
                temp = pd.DataFrame(value)
            attached_attributes = {}
            if self._structure is not None:
                for e in self.structure.dataset_attribute_codes:
                    if e in temp.keys():
                        if len(temp) > 0:
                            attached_attributes[e] = temp.loc[0, e]
                        del temp[e]

            self._data = temp.dropna(axis=1, how="all")
            if len(attached_attributes) > 0:
                for k, v in attached_attributes.items():
                    self.attached_attributes[k] = str(v)

    @property
    def dim_at_obs(self):
        """Extracts the dimensionAtObservation from the dataset_attributes"""
        return self.dataset_attributes.get('dimensionAtObservation')

    def read_csv(self, pathToCSV: str, **kwargs):
        """Loads the data from a CSV. Check the `Pandas read_csv docs
        <https://pandas.pydata.org/pandas-docs/stable/reference/api/pandas
        .read_csv.html>`_ Kwargs are supported

        :param pathToCSV: Path to CSV file
        :type pathToCSV: str

        """
        self._data = pd.read_csv(pathToCSV, **kwargs)

    def read_json(self, pathToJSON: str, **kwargs):
        """Loads the data from a JSON. Check the
        `Pandas read_json docs
        <https://pandas.pydata.org/pandas-docs/stable/reference/api/pandas
        .read_json.html>`_. Kwargs are supported

        :param pathToJSON: Path to JSON file
        :type pathToJSON: str
        """
        self._data = pd.read_json(pathToJSON, **kwargs)

    def read_excel(self, pathToExcel: str, **kwargs):
        """Loads the data from a Excel file. Check the `Pandas read_excel
        docs <https://pandas.pydata.org/pandas-docs/stable/reference/api
        /pandas.read_excel.html>`_. Kwargs are supported

        :param pathToExcel: Path to Excel file
        :type pathToExcel: str
        """
        self._data = pd.read_excel(pathToExcel, **kwargs)

    def to_csv(self, pathToCSV: str = None, **kwargs):
        """Parses the data to a CSV file. Kwargs are supported

        :param pathToCSV: Path to save as CSV file
        :type pathToCSV: str

        """
        return self.data.to_csv(pathToCSV, **kwargs)

    def to_json(self, pathToJSON: str = None):
        """Parses the data using the JSON Specification from the library
        documentation

        :param pathToJSON: Path to save as JSON file
        :type pathToJSON: str

        """

        element = {'structureRef': {"code": self.structure.id,
                                    "version": self.structure.version,
                                    "agencyID": self.structure.agencyID}}
        if len(self.dataset_attributes) > 0:
            element['dataset_attributes'] = self.dataset_attributes

        if len(self.attached_attributes) > 0:
            element['attached_attributes'] = self.attached_attributes

        result = self.data.to_json(orient="records")
        element['data'] = json.loads(result).copy()
        if pathToJSON is None:
            return element
        with open(pathToJSON, 'w') as f:
            f.write(json.dumps(element, ensure_ascii=False, indent=2))

    def to_feather(self, pathToFeather: str, **kwargs):
        """Parses the data to an Apache Feather format. Kwargs are supported.

        :param pathToFeather: Path to Feather file
        :type pathToFeather: str

        """
        self.data.to_feather(pathToFeather, **kwargs)

    def semantic_validation(self):
        """Performs a Semantic Validation on the Data.

        :returns:
            A list of errors as defined in the Validation Page.

        """
        if isinstance(self.data, DataFrame):
            return validate_data(self.data, self.structure)
        raise ValueError(f'Data for dataset {self.structure.id} '
                         f'is not well formed')

    def set_dimension_at_observation(self, dimAtObs):
        """Sets the dimensionAtObservation
            :param dimAtObs: Dimension At Observation
            :type dimAtObs: str
        """
        if (dimAtObs in self.structure.dimension_codes or
                dimAtObs == 'AllDimensions'):
            self.dataset_attributes['dimensionAtObservation'] = dimAtObs
        else:
            raise ValueError(f'{dimAtObs} is not a dimension '
                             f'of dataset {self.structure.id}')

    def _check_DA_keys(self, attributes: dict):
        """Inputs default values to the dataset_attributes in case they are
        missing

        :param attributes: A dictionary with dataset_attributes
        :type attributes: dict
        """
        keys = ["reportingBegin", "reportingEnd", "dataExtractionDate",
                "validFrom", "validTo", "publicationYear",
                "publicationPeriod", "action", "setId",
                "dimensionAtObservation"]

        for spared_key in attributes.keys():
            if spared_key not in keys:
                attributes.pop(spared_key)

        for k in keys:
            if k not in attributes.keys():
                if k == "dataExtractionDate":
                    attributes[k] = date.today()
                elif k == "action":
                    attributes[k] = "Replace"
                elif k == "setId":
                    if self.structure is not None:
                        attributes[k] = self.structure.id
                    elif self.dataflow is not None:
                        attributes[k] = self.dataflow.structure.id
                elif k == "dimensionAtObservation":
                    attributes[k] = "AllDimensions"
                else:
                    attributes[k] = None

        self._dataset_attributes = attributes.copy()

    def to_xml(self,
               message_type: MessageTypeEnum =
               MessageTypeEnum.StructureDataSet,
               outputPath: str = '',
               header: Header = None,
               id_: str = 'test',
               test: str = 'true',
               prepared: datetime = None,
               sender: str = 'Unknown',
               receiver: str = 'Not_supplied',
               prettyprint=True):
        """Parses the data to SDMX-ML 2.1, specifying the Message_Type
        (StructureSpecific or Generic or Metadata)

        :param message_type: Format of the Message in SDMX-ML
        :type message_type: MessageTypeEnum

        :param outputPath: Path to save the file, defaults to ''
        :type outputPath: str

        :param prettyprint: Saves the file formatted to be human readable
        :type prettyprint: bool

        :param header: Header to be written, defaults to None
        :type header: Header

        .. important::

            If the header argument is not None, rest of the below arguments
            will not be used

        :param id_: ID of the Header, defaults to 'test'
        :type id_: str

        :param test: Mark as test file, defaults to 'true'
        :type test: str

        :param prepared: Datetime of the preparation of the Message, \
        defaults to current date and time
        :type prepared: datetime

        :param sender: ID of the Sender, defaults to 'Unknown'
        :type sender: str

        :param receiver: ID of the Receiver, defaults to 'Not_supplied'
        :type receiver: str

        :returns: StringIO object, if outputPath is ''
        """

        if prepared is None:
            prepared = datetime.now()

        if outputPath == '':
            return writer(path=outputPath, type_=message_type, payload=self,
                          id_=id_, test=test, header=header,
                          prepared=prepared, sender=sender, receiver=receiver)
        writer(path=outputPath, type_=message_type, payload=self, id_=id_,
               test=test, header=header,
               prepared=prepared, sender=sender, receiver=receiver,
               prettyprint=prettyprint)



"""
    Message file contains the Message class for the use of external assets
"""
from datetime import datetime
from io import StringIO
from typing import Dict

from sdmxthon.model.dataset import Dataset
from sdmxthon.model.header import Party, Sender, Header
from sdmxthon.parsers.write import writer
from sdmxthon.utils.enums import MessageTypeEnum


class Message:
    """ Message class holds the type of a SDMX Message, its payload and its
    header.

    :param message_type: Enumeration that withholds the Message type for \
    writing purposes
    :type message_type: `MessageTypeEnum`

    :param payload: Information stored in the message (Datasets or structures)
    :type payload: :doc:`Dataset<./dataset>`

    :param header: Header of the message.
    :type header: `Header`
    """

    def __init__(self, message_type: MessageTypeEnum,
                 payload: (Dict[str, dict], Dict[str, Dataset], Dataset),
                 header: Header = None):
        self._type = message_type
        self._payload = payload
        self._header = header

    def __eq__(self, other):
        if isinstance(other, Message):
            return (self.type == other.type and
                    self.payload == other.payload and
                    self.header == other.header)

    @property
    def type(self):
        """Enumeration that withholds the Message type for writing purposes.

        :class: `MessageTypeEnum`

        """
        return self._type

    @type.setter
    def type(self, value):
        self._type = value

    @property
    def payload(self):
        """Information stored in the message (DataSet or Structure). A
        dictionary of datasets could also be used.

        :class: :doc:`Dataset<./dataset>`
        :class: `Dict [str, Dataset]`
        :class: `Structure`

        """
        if self._payload is None:
            raise ValueError('No Payload found')
        return self._payload

    @payload.setter
    def payload(self, value):
        if not isinstance(value, (dict, Dataset)):
            raise TypeError('Payload must be a DataSet, '
                            'a dict of DataSet or Metadata')
        self._payload = value

    @property
    def content(self):
        """Returns the payload as a dict

        :class: `Dict`

        """
        if isinstance(self.payload, dict):
            return self.payload
        if isinstance(self.payload, Dataset):
            return {'datasets': {self.payload.unique_id, self.payload}}

        return {'datasets': self.payload}

    @property
    def header(self):
        """Header of the message.

        :class: `Header`

        """
        return self._header

    @header.setter
    def header(self, value):

        self._header = value

    def set_dimension_at_observation(self, dimAtObs):
        """Sets the dimensionAtObservation if the payload is formed by Datasets

        :param dimAtObs: Dimension At Observation
        :type dimAtObs: str

        """
        for e in self.payload.values():
            if isinstance(e, Dataset):
                e.set_dimension_at_observation(dimAtObs)

    def header_creation(self, id_: str, test: bool = False,
                        senderId: str = "Unknown",
                        receiverId: str = "not_supplied",
                        datetimeStr=''):
        """
            Creates the header for a Message

            :param id_: ID of the Header
            :type id_: str

            :param test: Mark as test file
            :type test: bool

            :param senderId: ID of the Sender
            :type senderId: str

            :param receiverId: ID of the Receiver
            :type receiverId: str

            :param datetimeStr: Datetime of the preparation of the Message.
            :type datetimeStr: str Format:  '%Y-%m-%dT%H:%M:%S'
        """
        header = Header(ID=id_)
        header.test = test
        if datetimeStr == '':
            header.prepared = datetime.now()
        else:
            header.prepared = header._gds_parse_datetime(datetimeStr)

        sender = Sender()
        sender.id_ = senderId

        receiver = Party()
        receiver.id_ = receiverId

        header.sender = sender
        header.receiver = receiver

        self.header = header

    def validate(self):
        """Performs the semantic validation if the Payload is all Datasets

        :raises:
            TypeError: if the payload is not a DataSet or a dict of DataSets
        """
        validations = {}
        if (isinstance(self.payload, dict) and
                all(isinstance(n, Dataset) for n in self.payload.values())):
            for e in self.payload.values():
                list_errors = e.semantic_validation()
                if len(list_errors) > 0:
                    validations[e.structure.id] = list_errors
            if len(validations) == 0:
                return None

            return validations
        elif isinstance(self.payload, Dataset):
            return self.payload.semantic_validation()
        else:
            raise TypeError('Wrong Payload. Must be of type '
                            'DataSet or a dict of DataSet')

    def to_xml(self, outputPath: str = '',
               header: Header = None,
               id_: str = 'test',
               test: str = 'true',
               prepared: datetime = None,
               sender: str = 'Unknown',
               receiver: str = 'Not_supplied',
               prettyprint=True) -> StringIO:
        """Exports its payload to a XML file in SDMX-ML 2.1 format

        :param outputPath: Path to save the file, defaults to ''
        :type outputPath: str

        :param prettyprint: Specifies if the output file is formatted
        :type prettyprint: bool

        :param header: Header to be written, defaults to None
        :type header: Header

        .. important::

            If the header argument is not None, rest of the below arguments
            will not be used

        :param id_: ID of the Header, defaults to 'test'
        :type id_: str

        :param test: Mark as test file, defaults to 'true'
        :type test: str

        :param prepared: Datetime of the preparation of the Message, \
        defaults to current date and time
        :type prepared: datetime

        :param sender: ID of the Sender, defaults to 'Unknown'
        :type sender: str

        :param receiver: ID of the Receiver, defaults to 'Not_supplied'
        :type receiver: str

        :returns:
            StringIO object, if outputPath is ''

        """

        if prepared is None:
            prepared = datetime.now()

        if outputPath == '':
            return writer(path=outputPath, type_=self.type,
                          payload=self.payload, id_=id_, test=test,
                          prepared=prepared, sender=sender, receiver=receiver,
                          header=header, prettyprint=prettyprint)
        writer(path=outputPath, type_=self.type, payload=self.payload,
               id_=id_, test=test, header=header,
               prepared=prepared, sender=sender, receiver=receiver,
               prettyprint=prettyprint)


if __name__ == '__main__':
    print('Don´t do that!!')



# Common
HEADER = 'Header'
DATASET = 'DataSet'
SERIES = 'Series'
OBS = 'Obs'
AGENCY_ID = 'agencyID'
ID = 'id'
VERSION = 'version'

# Structure Specific
STRSPE = 'StructureSpecificData'
GENERIC = 'GenericData'
SERIESKEY = 'SeriesKey'
ATTRIBUTES = 'Attributes'
VALUE = 'Value'
OBS_DIM = 'ObsDimension'
OBSVALUE = 'ObsValue'
OBSKEY = 'ObsKey'
# Header
STRREF = 'structureRef'
STRUCTURE = 'Structure'
STR_USAGE = 'StructureUsage'
STRID = 'structureID'
DIM_OBS = 'dimensionAtObservation'
ALL_DIM = "AllDimensions"
REF = 'Ref'

# Structures
# Common
NAME = 'Name'
DESC = 'Description'
LANG = 'lang'
XML_TEXT = '#text'
STR_URL = 'structureURL'
STR_URL_LOW = 'structureUrl'
SER_URL = 'serviceURL'
SER_URL_LOW = 'serviceUrl'
# General
ANNOTATIONS = 'Annotations'
STRUCTURES = 'Structures'
ORGS = 'OrganisationSchemes'
AGENCIES = 'AgencyScheme'
CODELISTS = 'Codelists'
CONCEPTS = 'Concepts'
DSDS = 'DataStructures'
DATAFLOWS = 'Dataflows'
CONSTRAINTS = 'Constraints'

# Individual
AGENCY = 'Agency'
CL = 'Codelist'
CODE = 'Code'
CS = 'ConceptScheme'
CS_LOW = 'concept_scheme'
CON = 'Concept'
DSD = 'DataStructure'

# DSD components
DSD_COMPS = 'DataStructureComponents'
CON_ID = 'ConceptIdentity'
CON_ID_LOW = 'concept_identity'
# Dimension
DIM_LIST = 'DimensionList'
DIM_LIST_LOW = 'dimension_list'
DIM = 'Dimension'
TIME_DIM = 'TimeDimension'
# Attribute
ATT_LIST = 'AttributeList'
ATT_LIST_LOW = 'attribute_list'
ATT = 'Attribute'
ATT_REL = 'AttributeRelationship'
AS_STATUS = 'assignmentStatus'
# Measure
ME_LIST = 'MeasureList'
ME_LIST_LOW = 'measure_list'
MEASURE = 'Measure'
PRIM_MEASURE = 'PrimaryMeasure'
# Group Dimension
GROUP = 'Group'
GROUP_DIM_LOW = 'group_dimension_descriptor'
GROUP_DIM = 'GroupDimension'
DIM_REF = 'DimensionReference'

# Dataflows
DF = 'Dataflow'

# Constraints
CON_CONS = 'ContentConstraint'
CONS_ATT = 'ConstraintAttachment'
CUBE_REGION = 'CubeRegion'
CONTENT_REGION = 'dataContentRegion'
KEY = 'Key'
KEY_VALUE = 'KeyValue'
DATA_KEY_SET = 'DataKeySet'
DATA_KEY_SET_LOW = 'dataKeySet'
INCLUDED = 'isIncluded'
INCLUDE = 'include'

# Annotation
ANNOTATION = 'Annotation'
ANNOTATION_TITLE = 'AnnotationTitle'
ANNOTATION_TYPE = 'AnnotationType'
ANNOTATION_TEXT = 'AnnotationText'
ANNOTATION_URL = 'AnnotationURL'

TITLE = 'title'
TYPE_ = 'type_'
TYPE = 'type'
TEXT = 'text'
URL = 'url'

# Representation
CORE_REP = 'CoreRepresentation'
CORE_REP_LOW = 'core_representation'
LOCAL_REP = 'LocalRepresentation'
LOCAL_REP_LOW = 'local_representation'
ENUM = 'Enumeration'
ENUM_FORMAT = 'EnumerationFormat'
TEXT_FORMAT = 'TextFormat'

# Facets
FACETS = 'facets'
TEXT_TYPE = 'textType'
TEXT_TYPE_LOW = 'text_type'

# Contact
CONTACT = 'Contact'
DEPARTMENT = 'Department'
ROLE = 'Role'
URI = 'URI'
EMAIL = 'Email'
X400 = 'X400'
TELEPHONE = 'Telephone'
FAX = 'Fax'

# Extras
MAINTAINER = 'maintainer'
XMLNS = 'xmlns'
COMPS = 'components'
PARENT = 'Parent'
PAR_ID = 'maintainableParentID'
PAR_VER = 'maintainableParentVersion'
REL_TO = 'relatedTo'
NO_REL = 'NoSpecifiedRelationship'

# Schemas
pathToSchema = 'schemas/SDMXMessage.xsd'
schema_root = 'http://www.sdmx.org/resources/sdmxml/schemas/v2_1/'
namespaces = {
    schema_root + 'message': None,
    schema_root + 'common': None,
    schema_root + 'structure': None,
    'http://www.w3.org/2001/XMLSchema-instance': 'xsi',
    'http://www.w3.org/XML/1998/namespace': None,
    schema_root + 'data/structurespecific': None,
    schema_root + 'data/generic': None}

# To exclude from attached_attributes
exc_attributes = [STRREF, 'action', 'dataScope', 'xsi:type', SERIES, OBS]



from typing import List

from sdmxthon.model.utils import FacetType, string_setter, FacetValueType


class Facet:
    """Defines the format for the content of the Component when reported in
    a data or metadata set. """

    def __init__(self, facetType: str = None, facetValue: str = None,
                 facetValueType: str = None):
        self.facet_type = facetType
        self.facet_value = facetValue
        self.facet_value_type = facetValueType

    def __str__(self):
        return f'<{self.facet_type} - {self.facet_value}>'

    def __unicode__(self):
        return f'<{self.facet_type} - {self.facet_value}>'

    def __repr__(self):
        return f'<{self.facet_type} - {self.facet_value}>'

    @property
    def facet_type(self):
        """A specific content type which is constrained by the FacetType
        enumeration """
        return self._facetType

    @property
    def facet_value(self):
        """The value of the Facet"""
        return self._facetValue

    @property
    def facet_value_type(self):
        """The format of the value of a Component when reported in a data or
        metadata set. This is constrained by the FacetValueType enumeration.
        """
        return self._facetValueType

    @facet_type.setter
    def facet_type(self, value):
        if isinstance(value, str) or value is None:
            if value in FacetType or value is None:
                self._facetType = value
            else:
                raise ValueError(f"The facet {value} is not recognised")
        else:
            raise ValueError("Facet dim_type should be of the str dim_type")

    @facet_value.setter
    def facet_value(self, value):
        self._facetValue = string_setter(value)

    @facet_value_type.setter
    def facet_value_type(self, value):
        if isinstance(value, str) or value is None:
            if value in FacetValueType or value is None:
                self._facetValueType = value
            else:
                raise ValueError(
                    f"The facet value dim_type {value} is not recognised")
        else:
            raise ValueError(
                "Facet value dim_type should be of the str dim_type")


class Representation(object):
    """The allowable value or format for Component or Concept"""

    def __init__(self, facets: List[Facet] = None, codelist=None,
                 concept_scheme=None, text_type=None):
        self.codelist = codelist
        self.concept_scheme = concept_scheme
        self._type = text_type
        self._facets = []
        if facets is not None:
            for f in facets:
                self.add_facet(f)

    def __eq__(self, other):
        if isinstance(other, Representation):
            return (self._codelist == other._codelist and
                    self._conceptScheme == other._conceptScheme and
                    self._type == other._type)

        return False

    def __str__(self):
        return f'<{self.__class__.__name__} - {self.facets} - {self.type_}>'

    def __unicode__(self):
        return f'<{self.__class__.__name__} - {self.facets} - {self.type_}>'

    def __repr__(self):
        return f'<{self.__class__.__name__} - {self.facets} - {self.type_}>'

    @property
    def facets(self):
        """list of Facets found"""
        return self._facets

    @property
    def codelist(self):
        """Reference to the codelist"""
        return self._codelist

    @property
    def concept_scheme(self):
        """Reference to the ConceptScheme (only in MeasureDimension)"""
        return self._conceptScheme

    @codelist.setter
    def codelist(self, value):
        self._codelist = value

    @concept_scheme.setter
    def concept_scheme(self, value):
        self._conceptScheme = value

    @facets.setter
    def facets(self, value):
        self._facets = value

    def add_facet(self, value):
        """Add a facet to the list"""
        self._facets.append(value)

    @property
    def type_(self):
        """Specifies the basic type of the component (String, BigInteger...)"""
        return self._type



"""
    Handlers file provide functions to make the code more readable
"""


def first_element_dict(obj_: dict):
    if len(obj_) != 0:
        values_view = obj_.values()
        value_iterator = iter(values_view)
        first_value = next(value_iterator)
        return first_value
    else:
        return None


def split_unique_id(obj_: str):
    data = obj_.split(':', 1)
    agencyID = data[0]
    data = data[1].split('(', 1)
    id = data[0]
    version = data[1].split(')', 1)[0]

    return agencyID, id, version


def get_outfile(obj_: dict, key='', indent=''):
    element = obj_.get(key) or []

    outfile = ''

    for i in element:
        outfile += indent + i

    return outfile


def export_intern_data(data: dict, indent: str):
    outfile = get_outfile(data, 'Annotations', indent)
    outfile += get_outfile(data, 'Name', indent)
    outfile += get_outfile(data, 'Description', indent)

    return outfile


def add_indent(indent: str):
    if indent == '':
        return ''
    else:
        indent += '\t'
        return indent


def add_list(element: any):
    if not isinstance(element, list):
        element = [element]
    return element


def unique_id(agencyID, id_, version):
    return f"{agencyID}:{id_}({version})"


def recursive_compare(d1, d2, level='root'):
    control = True
    if isinstance(d1, dict) and isinstance(d2, dict):
        if d1.keys() != d2.keys():
            control = False
            s1 = set(d1.keys())
            s2 = set(d2.keys())
            print('{:<20} + {} - {}'.format(level, s1 - s2, s2 - s1))
            common_keys = s1 & s2
        else:
            common_keys = set(d1.keys())

        for k in common_keys:
            control = recursive_compare(d1[k], d2[k],
                                        level='{}.{}'.format(level, k))

    elif isinstance(d1, list) and isinstance(d2, list):
        if len(d1) != len(d2):
            control = False
            print('{:<20} len1={}; len2={}'.format(level, len(d1), len(d2)))
        common_len = min(len(d1), len(d2))
        if common_len != 0 and isinstance(d1[0], str):
            d1 = sorted(d1)
            d2 = sorted(d2)

        for i in range(common_len):
            control = recursive_compare(d1[i], d2[i],
                                        level='{}[{}]'.format(level, i))

    else:
        if d1 != d2:
            control = False
            print('{:<20} {} != {}'.format(level, d1, d2))

    return control



"""
SDMX Base package.

Please refer to the package in the SDMX Information Model

"""

from datetime import datetime
from typing import List

from sdmxthon.model.utils import string_setter, bool_setter, generic_setter, \
    set_date_from_string
from sdmxthon.utils.mappings import commonAbbr


class LocalisedString(object):
    """The Localised String supports the representation
       of text in one locale (locale is similar to language but
       includes geographic variations such as Canadian
       French, US English etc.).
    """

    def __init__(self, locale: str = None, label: str = None,
                 content: str = None):
        """Inits LocalisedString with optional attributes."""

        self.label = label
        self._locale = locale
        self._content = content

    def __eq__(self, other):
        if isinstance(other, LocalisedString):
            return (self._locale == other._locale and
                    self._content == other._content and
                    self._label == other._label)
        return False

    @property
    def locale(self):
        """Locale is the name of the country. Obtained from the
        `ISO 639-1 code
        <https://en.wikipedia.org/wiki/List_of_ISO_639-1_codes>`_"""
        return self._locale

    @locale.setter
    def locale(self, value):
        self._locale = string_setter(value)

    @property
    def label(self):
        """Label present in attribute lang of the XML namespace.
        Uses an ISO 639-1 code to specify the country"""
        return self._label

    @label.setter
    def label(self, value):
        self._label = string_setter(value)

    @property
    def content(self):
        """Added attribute to the Information Model to specify
        the string itself"""
        return self._content

    @content.setter
    def content(self, value):
        self._content = string_setter(value)


class InternationalString(object):
    """The International String is a collection of Localised
       Strings and supports the representation of text in
       multiple locales.
    """

    def __init__(self, localisedStrings: List[LocalisedString] = None):
        """Inits InternationalString with optional attributes."""

        self._items = {}
        if localisedStrings is not None:
            for record in localisedStrings:
                self.addLocalisedString(record)

    def __eq__(self, other):
        if isinstance(other, InternationalString):
            return self._items == other._items
        return False

    @property
    def items(self):
        """Items of the InternationalString"""
        return self._items

    def addLocalisedString(self, localisedString: LocalisedString):
        """Adds a localisedString to the International String"""
        if not isinstance(localisedString, LocalisedString):
            raise TypeError("International strings can only get localised "
                            "strings as arguments")
        self._items[localisedString.label] = {
            'locale': localisedString.locale,
            'content': localisedString.content}

    def getLabels(self):
        """Gets the labels of the items present in the International String"""
        return set(self._items.keys())

    def __getitem__(self, key):
        return self._items[key]

    def __str__(self):
        return str(self._items)

    def _to_XML(self, name, prettyprint):

        if prettyprint:
            child1 = '\t'
        else:
            child1 = ''

        outfile = []
        if len(self._items) > 0:
            for label, v in self._items.items():
                text = v["content"].replace("&", "&amp;") \
                    .replace("<", "&lt;").replace(">", "&gt;")
                outfile.append(
                    f'{child1}<{name} xml:lang="{label}">'
                    f'{text}</{name}>')

        return outfile


class Annotation(object):
    """Additional descriptive information attached to an object"""

    def __init__(self, id_: str = None, title: str = None, type_: str = None,
                 url: str = None, text: InternationalString = None):
        self.id = id_
        self.title = title
        self.type = type_
        self.url = url
        self.text = text

    def __eq__(self, other):
        if isinstance(other, Annotation):
            return (self._id == other._id and self._title == other._title and
                    self._type == other._type and self._url == other._url and
                    self._text == other._text)

    @staticmethod
    def _factory(*args_, **kwargs_):
        """Factory Method of Annotation"""
        return Annotation(*args_, **kwargs_)

    @property
    def id(self):
        """Identifier for the Annotation. It can be used to disambiguate one
            Annotation from another where there are several Annotations
            for the same annotated object."""
        return self._id

    @property
    def title(self):
        """A title used to identify an annotation."""
        return self._title

    @property
    def type(self):
        """Specifies how the annotation is to be processed"""
        return self._type

    @property
    def url(self):
        """A link to external descriptive text"""
        return self._url

    @property
    def text(self):
        """An International String provides the multilingual text content of
        the annotation via this role"""
        return self._text

    @id.setter
    def id(self, value):
        self._id = string_setter(value)

    @title.setter
    def title(self, value):
        self._title = string_setter(value)

    @type.setter
    def type(self, value):
        self._type = string_setter(value)

    @url.setter
    def url(self, value):
        self._url = string_setter(value)

    @text.setter
    def text(self, value):
        self._text = value


class AnnotableArtefact(object):
    """Superclass of all artefacts. Contains the list of annotations."""

    def __init__(self, annotations: List[Annotation] = None):

        self._annotations = []

        if isinstance(annotations, list):
            for a in annotations:
                self.addAnnotation(a)
        elif annotations is None:
            self._annotations = []
        else:
            raise TypeError(
                "Annotations passed to Annotable artefacts can only be a list "
                "of Annotation objects or a single Annotation")

    def __eq__(self, other):
        if isinstance(other, AnnotableArtefact):
            return self._annotations == other._annotations

    @property
    def annotations(self):
        """List of the annotations of the element"""
        return self._annotations

    def addAnnotation(self, annotation: Annotation):
        """Method to add an annotation to the list"""
        if not isinstance(annotation, Annotation):
            raise TypeError("Annotable artefacts can only get "
                            "annotations as arguments")
        self._annotations.append(annotation)

    def _to_XML(self, prettyprint):

        if prettyprint:
            child1 = '\t'
            child2 = '\t\t'
        else:
            child1 = child2 = ''
        if len(self.annotations) > 0:
            outfile = [f'<{commonAbbr}:Annotations>']
            for e in self.annotations:
                if e.id is None:
                    outfile.append(f'{child1}<{commonAbbr}:Annotation>')
                else:
                    outfile.append(
                        f'{child1}<{commonAbbr}:Annotation id="{e.id}">')

                if e.title is not None and e.title != '':
                    outfile.append(
                        f'{child2}<{commonAbbr}:AnnotationTitle>'
                        f'{e.title.replace("&", "&amp;").rstrip()}'
                        f'</{commonAbbr}:AnnotationTitle>')

                if e.type is not None and e.type != '':
                    outfile.append(
                        f'{child2}<{commonAbbr}:AnnotationType>'
                        f'{e.type.replace("&", "&amp;").rstrip()}'
                        f'</{commonAbbr}:AnnotationType>')

                if e.url is not None and e.url != '':
                    outfile.append(
                        f'{child2}<{commonAbbr}:AnnotationURL>'
                        f'{e.url.replace("&", "&amp;")}'
                        f'</{commonAbbr}:AnnotationURL>')

                if e.text is not None:
                    outfile += [child1 + i for i in
                                e._text._to_XML(f"{commonAbbr}:AnnotationText",
                                                prettyprint)]
                outfile.append(f'{child1}</{commonAbbr}:Annotation>')
            outfile.append(f'</{commonAbbr}:Annotations>')
            return outfile

        return None


class IdentifiableArtefact(AnnotableArtefact):
    """Provides identity to all derived classes. It also provides annotations
    to derived classes because it is a subclass of AnnotableArtefact.
    """

    def __init__(self, id_: str = None, uri: str = None, urn: str = None,
                 annotations: List[Annotation] = None):

        super(IdentifiableArtefact, self).__init__(annotations=annotations)

        self.id = id_
        self.uri = uri
        self.urn = urn

    def __eq__(self, other):
        if isinstance(other, IdentifiableArtefact):
            return (super(IdentifiableArtefact, self).__eq__(other) and
                    self._id == other._id and
                    self._uri == other._uri and self._urn == other._urn)
        return False

    @property
    def id(self):
        """The unique identifier of the object."""
        return self._id

    @property
    def uri(self):
        """Universal resource identifier that may or may not be resolvable."""
        return self._uri

    @property
    def urn(self):
        """Universal resource name – this is for use in registries: all
        registered objects have a urn."""
        return self._urn

    @id.setter
    def id(self, value):
        self._id = value

    @uri.setter
    def uri(self, value):
        self._uri = value

    @urn.setter
    def urn(self, value):
        self._urn = generic_setter(value, str)

    def _to_XML(self, prettyprint) -> dict:

        attributes = ''

        if self.id is not None:
            attributes += f' id="{self.id}"'

        if self.uri is not None:
            attributes += f' uri="{self.uri}"'

        if self.urn is not None:
            attributes += f' urn="{self.urn}"'

        outfile = {'Annotations': super(IdentifiableArtefact, self)._to_XML(
            prettyprint), 'Attributes': attributes}

        return outfile


class NameableArtefact(IdentifiableArtefact):
    """Provides a Name and Description to all derived classes in
    addition to identification and annotations."""

    def __init__(self, id_: str = None, uri: str = None, urn=None,
                 annotations: List[Annotation] = None,
                 name: InternationalString = None,
                 description: InternationalString = None):
        super(NameableArtefact, self).__init__(id_=id_, uri=uri, urn=urn,
                                               annotations=annotations)

        self._name = name
        self._description = description

    def __eq__(self, other):
        if isinstance(other, NameableArtefact):
            return (super(NameableArtefact, self).__eq__(other) and
                    self._name == other._name and
                    self._description == other._description)
        return False

    @property
    def name(self):
        """A multi-lingual name is provided by this role via
        the International String class """
        if self._name is None:
            return None
        if isinstance(self._name, InternationalString):
            if len(self._name.items) == 0:
                return None
            if len(self._name.items) == 1:
                values_view = self._name.items.values()
                value_iterator = iter(values_view)
                first_value = next(value_iterator)
                return first_value['content']

            return self._name.items

        return self._name

    @property
    def description(self):
        """A multi-lingual description is provided by this role
        via the International String class."""
        if self._description is None:
            return None
        if isinstance(self._description, InternationalString):
            if len(self._description.items) == 0:
                return None

            if len(self._description.items) == 1:
                values_view = self._description.items.values()
                value_iterator = iter(values_view)
                first_value = next(value_iterator)
                return first_value['content']

            return self._description.items
        return self._description

    @name.setter
    def name(self, value):
        self._name = generic_setter(value, InternationalString)

    @description.setter
    def description(self, value):
        self._description = generic_setter(value, InternationalString)

    def _to_XML(self, prettyprint) -> dict:
        outfile = super(NameableArtefact, self)._to_XML(prettyprint)

        if self.name is not None:
            outfile['Name'] = self._name._to_XML(f'{commonAbbr}:Name',
                                                 prettyprint)

        if self._description is not None:
            outfile['Description'] = self._description._to_XML(
                f'{commonAbbr}:Description', prettyprint)

        return outfile


class VersionableArtefact(NameableArtefact):
    """Provides versioning information for all derived objects"""

    def __init__(self, id_: str, uri: str = None, urn=None,
                 annotations: List[Annotation] = None,
                 name: str = None, description: str = None,
                 version: str = "1.0", validFrom: datetime = None,
                 validTo: datetime = None):

        super(VersionableArtefact, self).__init__(id_=id_, uri=uri, urn=urn,
                                                  annotations=annotations,
                                                  name=name,
                                                  description=description)
        self.version = version
        self.validFrom = validFrom
        self.validTo = validTo

    def __eq__(self, other):
        if isinstance(other, VersionableArtefact):
            return (super(VersionableArtefact, self).__eq__(other) and
                    self._version == other._version and
                    self._validFrom == other._validFrom and
                    self._validTo == other._validTo)

        return False

    @property
    def version(self):
        """A version string following an agreed convention"""
        return self._version

    @property
    def validFrom(self):
        """Date from which the version is valid"""
        return self._validFrom

    @property
    def validTo(self):
        """Date from which version is superseded"""
        return self._validTo

    @version.setter
    def version(self, value):
        self._version = string_setter(value, pattern="[0-9]+(.[0-9]+)*")

    @validFrom.setter
    def validFrom(self, value):
        self._validFrom = set_date_from_string(value)

    @validTo.setter
    def validTo(self, value):
        self._validTo = set_date_from_string(value)

    def _to_XML(self, prettyprint) -> dict:
        outfile = super(VersionableArtefact, self)._to_XML(prettyprint)

        if self.version is not None:
            outfile['Attributes'] += f' version="{self.version}"'

        if self.validFrom is not None:
            valid_from_str = self.validFrom.strftime("%Y-%m-%dT%H:%M:%S")
            outfile['Attributes'] += f' validFrom="{valid_from_str}"'

        if self.validTo is not None:
            valid_to_str = self.validTo.strftime("%Y-%m-%dT%H:%M:%S")
            outfile['Attributes'] += f' validTo="{valid_to_str}"'

        return outfile


class MaintainableArtefact(VersionableArtefact):
    """An abstract class to group together primary structural metadata artefacts
       that are maintained by an Agency."""

    def __init__(self, id_: str = None, uri: str = None, urn=None,
                 annotations: List[Annotation] = None,
                 name: str = None, description: str = None,
                 version: str = None, validFrom: datetime = None,
                 validTo: datetime = None,
                 isFinal: bool = False, isExternalReference: bool = False,
                 serviceUrl: str = None,
                 structureUrl: str = None, maintainer=None):
        super(MaintainableArtefact, self).__init__(id_=id_, uri=uri, urn=urn,
                                                   annotations=annotations,
                                                   name=name,
                                                   description=description,
                                                   version=version,
                                                   validFrom=validFrom,
                                                   validTo=validTo)

        self.isFinal = isFinal
        self.isExternalReference = isExternalReference
        self.serviceUrl = serviceUrl
        self.structureUrl = structureUrl
        self.maintainer = maintainer
        if id_ is not None and maintainer is not None and version is not None:
            self._unique_id = f'{maintainer.id}:{id_}({version})'
        else:
            self._unique_id = None

    def __eq__(self, other):
        if isinstance(other, MaintainableArtefact):
            return (super(MaintainableArtefact, self).__eq__(other) and
                    self._isFinal == other._isFinal and
                    self._isExternalReference == other._isExternalReference and
                    self._serviceUrl == other._serviceUrl and
                    self._structureUrl == other._structureUrl and
                    self._maintainer == other._maintainer)
        return False

    @property
    def unique_id(self):
        """Provides the unique id in the shape of AgencyID:ID(version)"""
        if isinstance(self.maintainer, str):
            return f'{self.maintainer}:{self.id}({self.version})'

        return f'{self.maintainer.id}:{self.id}({self.version})'

    @property
    def isFinal(self):
        """Defines whether a maintained artefact is draft or final"""
        return self._isFinal

    @property
    def isExternalReference(self):
        """If set to “true” it indicates that the content of the object
        is held externally."""
        return self._isExternalReference

    @property
    def serviceUrl(self):
        """The URL of an SDMX compliant web service from which the external
        object can be retrieved."""
        return self._serviceUrl

    @property
    def structureUrl(self):
        """The URL of an SDMX-ML document containing the external object."""
        return self._structureUrl

    @property
    def maintainer(self):
        """Association to the Maintenance Agency responsible for maintaining
        the artefact."""
        return self._maintainer

    @isFinal.setter
    def isFinal(self, value):
        self._isFinal = bool_setter(value)

    @isExternalReference.setter
    def isExternalReference(self, value):
        self._isExternalReference = bool_setter(value)

    @serviceUrl.setter
    def serviceUrl(self, value):
        self._serviceUrl = string_setter(value)

    @structureUrl.setter
    def structureUrl(self, value):
        self._structureUrl = string_setter(value)

    @maintainer.setter
    def maintainer(self, value):
        self._maintainer = value

    @property
    def agencyID(self):
        """Extracts the agencyID from the maintainer"""
        if self.maintainer is not None:
            if isinstance(self.maintainer, str):
                return self.maintainer

            return self.maintainer.id

        return None

    def _to_XML(self, prettyprint):
        outfile = super(MaintainableArtefact, self)._to_XML(prettyprint)

        if self.isExternalReference is not None:
            outfile['Attributes'] += \
                f' isExternalReference="' \
                f'{str(self.isExternalReference).lower()}' \
                f'"'

        if self.isFinal is not None:
            outfile['Attributes'] += f' isFinal="{str(self.isFinal).lower()}"'

        if self.maintainer is not None:
            if isinstance(self.maintainer, str):
                outfile['Attributes'] += f' agencyID="{self.maintainer}"'
            else:
                outfile['Attributes'] += f' agencyID="{self.maintainer.id}"'

        return outfile



import copy

from sdmxthon.model.base import InternationalString, LocalisedString, \
    Annotation
from sdmxthon.model.component import Component, Dimension, TimeDimension, \
    Attribute, PrimaryMeasure
from sdmxthon.model.definitions import DataStructureDefinition, \
    DataFlowDefinition, ContentConstraint, DataKeySet, \
    MemberSelection, CubeRegion
from sdmxthon.model.descriptors import DimensionDescriptor, \
    AttributeDescriptor, MeasureDescriptor, GroupDimensionDescriptor
from sdmxthon.model.header import Contact
from sdmxthon.model.itemScheme import Agency, AgencyScheme, Codelist, Code, \
    Item, ConceptScheme, Concept
from sdmxthon.model.representation import Representation, Facet
from sdmxthon.model.utils import FacetType
from sdmxthon.utils.handlers import add_list, unique_id
from sdmxthon.utils.mappings import Locale_Codes
from sdmxthon.utils.parsing_words import ORGS, AGENCIES, AGENCY, ID, \
    AGENCY_ID, VERSION, NAME, DESC, LANG, XML_TEXT, URI, EMAIL, ROLE, \
    DEPARTMENT, TELEPHONE, FAX, CONTACT, MAINTAINER, CODELISTS, CL, \
    CODE, CONCEPTS, CS, CON, ANNOTATIONS, ANNOTATION, ANNOTATION_TITLE, \
    TITLE, ANNOTATION_TYPE, TYPE_, ANNOTATION_TEXT, TEXT, CORE_REP, \
    CORE_REP_LOW, ENUM, REF, XMLNS, ENUM_FORMAT, TEXT_FORMAT, \
    TEXT_TYPE, TEXT_TYPE_LOW, FACETS, DSDS, DSD, DSD_COMPS, DIM_LIST, \
    ATT_LIST, ME_LIST, GROUP, DIM_LIST_LOW, ATT_LIST_LOW, ME_LIST_LOW, DIM, \
    TIME_DIM, ATT, COMPS, CON_ID, PAR_ID, PAR_VER, CS_LOW, LOCAL_REP, \
    LOCAL_REP_LOW, ATT_REL, REL_TO, PRIM_MEASURE, DATAFLOWS, ANNOTATION_URL, \
    URL, CON_ID_LOW, PARENT, GROUP_DIM_LOW, GROUP_DIM, DIM_REF, DF, \
    STRUCTURE, STR_URL, STR_URL_LOW, SER_URL, SER_URL_LOW, CONSTRAINTS, \
    CON_CONS, CONS_ATT, DATA_KEY_SET, DATA_KEY_SET_LOW, CUBE_REGION, \
    KEY_VALUE, KEY, VALUE, TYPE, INCLUDED, INCLUDE, CONTENT_REGION

schemes_classes = {CL: Codelist, AGENCIES: AgencyScheme, CS: ConceptScheme}
items_classes = {AGENCY: Agency, CODE: Code, CON: Concept}
comp_lists_classes = {DIM_LIST: DimensionDescriptor,
                      ATT_LIST: AttributeDescriptor,
                      ME_LIST: MeasureDescriptor,
                      GROUP: GroupDimensionDescriptor}

comp_classes = {DIM: Dimension,
                TIME_DIM: TimeDimension,
                ATT: Attribute,
                PRIM_MEASURE: PrimaryMeasure}

comp_lists_names = {DIM_LIST: DIM_LIST_LOW,
                    ATT_LIST: ATT_LIST_LOW,
                    ME_LIST: ME_LIST_LOW}

comp_lists_items = {DIM_LIST: [DIM, TIME_DIM],
                    ATT_LIST: [ATT],
                    ME_LIST: [PRIM_MEASURE]}

dimensions = {}
measures = {}
groups = {}

# Global dict to be used in all elements
agencies = {}
codelists = {}
concepts = {}
datastructures = {}
dataflows = {}

# Errors
errors = []
missing_rep = {"CON": [], "CS": [], "CL": []}
dsd_id = ""


def create_int_str(json_int) -> InternationalString:
    json_int = add_list(json_int)
    locals_list = []
    for e in json_int:
        if e[XML_TEXT].strip() not in ['', '\n']:
            e[XML_TEXT] = " ".join(e[XML_TEXT].split())
        locals_list.append(LocalisedString(locale=Locale_Codes[e[LANG]],
                                           label=e[LANG],
                                           content=e[XML_TEXT]))
    return InternationalString(localisedStrings=locals_list)


def create_contact(json_contact) -> Contact:
    node_int = [NAME, DEPARTMENT, ROLE]
    node_str = [URI, EMAIL, TELEPHONE, FAX]

    for e in node_int:
        if e in json_contact:
            json_contact[e.lower()] = create_int_str(json_contact[e])
            del json_contact[e]
    for e in node_str:
        if e in json_contact:
            json_contact[e.lower()] = add_list(json_contact.pop(e))

    return Contact(**json_contact)


def format_name_description(element: any):
    node = [NAME, DESC]
    for e in node:
        if e in element:
            element[e.lower()] = create_int_str(element[e])
            del element[e]
    return element


def format_id(element: any):
    element[ID + '_'] = element.pop(ID)
    return element


def format_maintainer(element: any):
    if element[AGENCY_ID] in agencies:
        element[MAINTAINER] = agencies[element[AGENCY_ID]]
    else:
        element[MAINTAINER] = Agency(element[AGENCY_ID])
    del element[AGENCY_ID]
    return element


def format_annotations(item_elem: any):
    annotations = []
    if ANNOTATIONS in item_elem:
        ann = item_elem[ANNOTATIONS]
        if ANNOTATION in ann:
            ann[ANNOTATION] = add_list(ann[ANNOTATION])
            for e in ann[ANNOTATION]:
                if ANNOTATION_TITLE in e:
                    e[TITLE] = e.pop(ANNOTATION_TITLE)
                if ANNOTATION_TYPE in e:
                    e[TYPE_] = e.pop(ANNOTATION_TYPE)
                if ANNOTATION_TEXT in e:
                    e[TEXT] = create_int_str(e[ANNOTATION_TEXT])
                    del e[ANNOTATION_TEXT]
                if ANNOTATION_URL in e:
                    e[URL] = e.pop(ANNOTATION_URL)

                annotations.append(Annotation(**e))

        item_elem[ANNOTATIONS.lower()] = annotations
        del item_elem[ANNOTATIONS]

    return item_elem


def format_facets(json_fac) -> dict:
    fac = {FACETS: []}
    if json_fac is None:
        return fac
    if TEXT_TYPE in json_fac:
        fac[TEXT_TYPE_LOW] = json_fac.pop(TEXT_TYPE)
    for e in json_fac:
        if e in FacetType:
            fac[FACETS].append(Facet(facetType=e, facetValue=json_fac[e]))

    return fac


def format_representation(json_rep) -> Representation:
    rep = {}
    node = [ENUM_FORMAT, TEXT_FORMAT]
    if ENUM in json_rep:
        data = json_rep[ENUM][REF]
        full_id = unique_id(data[AGENCY_ID], data[ID], data[VERSION])
        if full_id in codelists:
            rep[CL.lower()] = codelists[full_id]
        elif full_id not in missing_rep["CL"]:
            missing_rep["CL"].append(full_id)
            rep[CL.lower()] = full_id

    for e in node:
        if e in json_rep:
            rep = {**rep, **format_facets(json_rep[e])}

    return Representation(**rep)


def format_urls(json_elem):
    if STR_URL in json_elem:
        json_elem[STR_URL_LOW] = json_elem.pop(STR_URL)
    if SER_URL in json_elem:
        json_elem[SER_URL_LOW] = json_elem.pop(SER_URL)
    return json_elem


def create_item(item_elem, item) -> Item:
    if XMLNS in item_elem:
        del item_elem[XMLNS]

    item_elem = format_annotations(item_elem)
    item_elem = format_name_description(item_elem)
    item_elem = format_id(item_elem)

    if CONTACT in item_elem and item == AGENCY:
        item_elem[CONTACT] = add_list(item_elem[CONTACT])
        contacts = []
        for e in item_elem[CONTACT]:
            contacts.append(create_contact(e))
        item_elem[CONTACT.lower() + 's'] = contacts
        del item_elem[CONTACT]

    if CORE_REP in item_elem and item == CON:
        item_elem[CORE_REP_LOW] = format_representation(item_elem[CORE_REP])
        del item_elem[CORE_REP]

    if PARENT in item_elem:
        item_elem[PARENT.lower()] = item_elem.pop(PARENT)[REF][ID]

    return items_classes[item](**item_elem)


def create_scheme(json_elem, scheme, item):
    elements = {}

    if scheme in json_elem:
        json_elem[scheme] = add_list(json_elem[scheme])
        for element in json_elem[scheme]:

            if XMLNS in element:
                del element[XMLNS]

            element = format_annotations(element)
            element = format_name_description(element)
            full_id = unique_id(element[AGENCY_ID],
                                element[ID],
                                element[VERSION])
            element = format_urls(element)
            element = format_maintainer(element)
            element = format_id(element)
            if item in element:
                element[item] = add_list(element[item])
                items = []
                for item_elem in element[item]:
                    # Dynamic
                    items.append(create_item(item_elem, item))
                del element[item]
                element['items'] = items
                if scheme == AGENCIES:
                    agencies.update({e.id: e for e in items})
            else:
                element['items'] = []
            # Dynamic creation with specific class
            elements[full_id] = schemes_classes[scheme](**element)

    return elements


def create_organisations(json_orgs):
    orgs = {}
    if AGENCIES in json_orgs:
        if len(json_orgs) == 1 and isinstance(json_orgs[AGENCIES], dict):
            ag_sch = create_scheme(json_orgs, AGENCIES, AGENCY)
            return ag_sch
        for e in json_orgs[AGENCIES]:
            ag_sch = create_scheme(e, AGENCIES, AGENCY)
            orgs = {**orgs, **ag_sch}
    return orgs


def format_con_id(json_ref):
    rep = {}
    full_cs_id = unique_id(json_ref[AGENCY_ID],
                           json_ref[PAR_ID],
                           json_ref[PAR_VER])

    if full_cs_id in concepts:
        if json_ref[ID] in concepts[full_cs_id].items:
            # rep[CS_LOW] = concepts[full_cs_id]
            rep[CON] = concepts[full_cs_id].items[json_ref[ID]]
            core_rep = concepts[full_cs_id].items[json_ref[ID]]. \
                core_representation
            if core_rep is not None:
                cl = core_rep.codelist
                if cl is not None:
                    rep[CL.lower()] = cl
        elif json_ref[ID] not in missing_rep["CON"]:
            missing_rep["CON"].append(json_ref[ID])

    elif full_cs_id not in missing_rep["CS"]:
        missing_rep["CS"].append(full_cs_id)

    return rep


def format_relationship(json_rel, node=DIM, att_name=None):
    rels = {}
    if node in json_rel:
        json_rel[node] = add_list(json_rel[node])
        for e in json_rel[node]:
            if DIM_REF in e:
                element = e[DIM_REF][REF][ID]
            else:
                element = e[REF][ID]
            if element in dimensions:
                rels[element] = dimensions[element]
            else:
                errors.append(
                    {'Code': 'MS04', 'ErrorLevel': 'CRITICAL',
                     'ObjectID': f'{dsd_id}',
                     'ObjectType': 'Attribute',
                     'Message': f'Missing Dimension {e[REF][ID]} '
                                f'related to Attribute '
                                f'{att_name}'})
    elif PRIM_MEASURE in json_rel:
        if json_rel[PRIM_MEASURE][REF][ID] in measures:
            rels = measures[json_rel[PRIM_MEASURE][REF][ID]]
        else:
            errors.append(
                {'Code': 'MS05', 'ErrorLevel': 'CRITICAL',
                 'ObjectID': f'{dsd_id}',
                 'ObjectType': 'Attribute',
                 'Message': 'Missing Primary Measure '
                            f'{json_rel[PRIM_MEASURE][REF][ID]} '
                            f'related to Attribute {att_name}'})
    elif GROUP in json_rel:
        if json_rel[GROUP][REF][ID] in groups:
            rels = groups[json_rel[GROUP][REF][ID]]
    else:
        rels = 'NoSpecifiedRelationship'
    return rels


def format_component(json_comp, comp) -> Component:
    rep = {}
    rep_class = None
    if LOCAL_REP in json_comp:
        rep_class = format_representation(json_comp[LOCAL_REP])
        del json_comp[LOCAL_REP]
    if CON_ID in json_comp:
        rep = format_con_id(json_comp[CON_ID][REF])
        if CON in rep:
            json_comp[CON_ID_LOW] = rep.pop(CON)
        del json_comp[CON_ID]
    if CS_LOW in rep:
        rep_class.concept_scheme = rep[CS_LOW]

    json_comp[LOCAL_REP_LOW] = rep_class

    # Attribute Handling
    if ATT_REL in json_comp:
        json_comp[REL_TO] = format_relationship(json_comp[ATT_REL],
                                                att_name=json_comp[ID])
        del json_comp[ATT_REL]
    json_comp = format_id(json_comp)
    json_comp = format_annotations(json_comp)

    return comp(**json_comp)


def format_component_lists(json_comp_lists, comp_list, comp):
    components = {}
    json_comp_lists = format_annotations(json_comp_lists)
    json_comp_lists = format_id(json_comp_lists)
    for e in comp:
        if e in json_comp_lists:
            json_comp_lists[e] = add_list(json_comp_lists[e])
            for i in json_comp_lists[e]:
                new_element = format_component(i, comp_classes[e])
                components[new_element.id] = new_element
            del json_comp_lists[e]

    if comp == comp_lists_items[DIM_LIST]:
        dimensions.update(components)
    elif comp == comp_lists_items[ME_LIST]:
        measures.update(components)
        if len(components) == 0:
            errors.append({'Code': 'MX02', 'ErrorLevel': 'CRITICAL',
                           'ObjectID': f'{dsd_id}', 'ObjectType': 'DSD',
                           'Message': f'DSD {dsd_id} does not have '
                                      f'a Primary Measure'})

    json_comp_lists[COMPS] = components

    return comp_list(**json_comp_lists)


def format_group_dim(json_group):
    if GROUP_DIM in json_group:
        json_group[COMPS] = format_relationship(json_group, GROUP_DIM)
    else:
        json_group[COMPS] = None

    del json_group[GROUP_DIM]
    groups.update({json_group[ID]: json_group[COMPS]})
    json_group = format_id(json_group)

    return GroupDimensionDescriptor(**json_group)


def format_dsd_comps(json_comps):
    node = [DIM_LIST, ME_LIST, GROUP, ATT_LIST]
    comps = json_comps[DSD_COMPS]
    for e in node:
        if e == GROUP and e in comps:
            json_comps[GROUP_DIM_LOW] = format_group_dim(comps[GROUP])
        elif e in comps:
            name = comp_lists_names[e]
            json_comps[name] = format_component_lists(comps[e],
                                                      comp_lists_classes[e],
                                                      comp_lists_items[e])
        else:
            if e == DIM_LIST:
                errors.append({'Code': 'MX01', 'ErrorLevel': 'CRITICAL',
                               'ObjectID': f'{dsd_id}', 'ObjectType': 'DSD',
                               'Message': f'DSD {dsd_id} does not have '
                                          f'a DimensionList'})
            elif e == ME_LIST:
                errors.append({'Code': 'MX02', 'ErrorLevel': 'CRITICAL',
                               'ObjectID': f'{dsd_id}', 'ObjectType': 'DSD',
                               'Message': f'DSD {dsd_id} does not have '
                                          f'a Primary Measure'})
    del json_comps[DSD_COMPS]

    return json_comps


def create_datastructures(json_dsds):
    elements = {}
    if json_dsds is not None and DSD in json_dsds:
        json_dsds[DSD] = add_list(json_dsds[DSD])
        for element in json_dsds[DSD]:
            dimensions.clear()
            measures.clear()
            groups.clear()
            element = format_annotations(element)
            element = format_name_description(element)
            full_id = unique_id(element[AGENCY_ID],
                                element[ID],
                                element[VERSION])
            global dsd_id
            dsd_id = full_id
            element = format_urls(element)

            element = format_maintainer(element)
            element = format_id(element)
            if XMLNS in element:
                del element[XMLNS]

            if DSD_COMPS in element:
                element = format_dsd_comps(element)
            del dsd_id
            # Creation of DSD
            if full_id not in elements:
                elements[full_id] = DataStructureDefinition(**element)
            else:
                errors.append({'Code': 'MS06', 'ErrorLevel': 'CRITICAL',
                               'ObjectID': f'{full_id}', 'ObjectType': 'DSD',
                               'Message': f'DSD {full_id} is not unique'})

    else:
        errors.append(
            {'Code': 'MS01', 'ErrorLevel': 'CRITICAL',
             'ObjectID': None,
             'ObjectType': 'DSD',
             'Message': 'Not found any DSD in this file'})
    return elements


def create_dataflows(json_dfs):
    elements = {}
    if DF in json_dfs:
        json_dfs[DF] = add_list(json_dfs[DF])
        for element in json_dfs[DF]:
            if XMLNS in element:
                del element[XMLNS]

            element = format_annotations(element)
            element = format_name_description(element)

            if STRUCTURE in element:
                agency_id = element[STRUCTURE][REF][AGENCY_ID]
                id_ = element[STRUCTURE][REF][ID]
                version = element[STRUCTURE][REF][VERSION]

                str_id = unique_id(agency_id, id_, version)

                if str_id in datastructures:
                    element[STRUCTURE.lower()] = datastructures[str_id]

                del element[STRUCTURE]

            full_id = unique_id(element[AGENCY_ID],
                                element[ID],
                                element[VERSION])
            element = format_urls(element)

            element = format_maintainer(element)
            element = format_id(element)
            # Creation of DSD
            if full_id not in elements:
                elements[full_id] = DataFlowDefinition(**element)
            else:
                raise Exception

    return elements


def format_key_set(json_key_set):
    json_key_set[KEY_VALUE] = add_list(json_key_set[KEY_VALUE])
    key_set = {}
    for e in json_key_set[KEY_VALUE]:
        key_set[e[ID]] = e[VALUE]

    return key_set


def format_restrictions(json_cons) -> dict:
    if DATA_KEY_SET in json_cons:
        json_cons[DATA_KEY_SET] = add_list(json_cons[DATA_KEY_SET])
        json_cons[DATA_KEY_SET_LOW] = []
        for element in json_cons[DATA_KEY_SET]:
            list_keys = []
            element[KEY] = add_list(element[KEY])
            for e in element[KEY]:
                list_keys.append(format_key_set(e))

            json_cons[DATA_KEY_SET_LOW].append(DataKeySet(keys=list_keys,
                                                          isIncluded=element[
                                                              INCLUDED]))
        del json_cons[DATA_KEY_SET]
    if CUBE_REGION in json_cons:
        json_cons[CUBE_REGION] = add_list(json_cons[CUBE_REGION])
        cubes = []
        for element in json_cons[CUBE_REGION]:

            is_included = element[INCLUDE]
            keys = format_key_set(element)
            members = []
            for e in keys:
                members.append(MemberSelection(is_included,
                                               values_for=e,
                                               sel_value=keys[e]))

            cubes.append(CubeRegion(is_included=is_included, member=members))

        json_cons[CONTENT_REGION] = cubes
        del json_cons[CUBE_REGION]

    return json_cons


def create_constraints(json_cons):
    elements = {}
    if CON_CONS in json_cons:
        json_cons[CON_CONS] = add_list(json_cons[CON_CONS])
        for element in json_cons[CON_CONS]:
            attachment = None
            references = []
            if XMLNS in element:
                del element[XMLNS]

            element = format_annotations(element)
            element = format_name_description(element)
            full_id = unique_id(element[AGENCY_ID],
                                element[ID],
                                element[VERSION])
            element = format_urls(element)

            element = format_maintainer(element)
            element = format_id(element)

            if CONS_ATT in element:
                if DSD in element[CONS_ATT]:
                    agency_id = element[CONS_ATT][DSD][REF][AGENCY_ID]
                    id_ = element[CONS_ATT][DSD][REF][ID]
                    version = element[CONS_ATT][DSD][REF][VERSION]

                    str_id = unique_id(agency_id, id_, version)

                    references = [str_id, DSD]

                    if str_id in datastructures:
                        attachment = datastructures[str_id]

                elif DF in element[CONS_ATT]:
                    agency_id = element[CONS_ATT][DF][REF][AGENCY_ID]
                    id_ = element[CONS_ATT][DF][REF][ID]
                    version = element[CONS_ATT][DF][REF][VERSION]

                    str_id = unique_id(agency_id, id_, version)

                    references = [str_id, DF]

                    if str_id in dataflows:
                        attachment = dataflows[str_id]
                del element[CONS_ATT]

            if TYPE in element:
                element[ROLE.lower()] = element.pop(TYPE)

            element = format_restrictions(element)

            # Creation of Constraint
            if full_id not in elements:
                elements[full_id] = ContentConstraint(**element)
            else:
                raise Exception

            if attachment is not None:
                attachment.add_constraint(elements[full_id])
                del attachment

            # TODO Delete once we change constraints

            if len(references) > 0:
                elements[full_id]._ref_attach = references[0]
                elements[full_id]._type_attach = references[1]

    return elements


def grouping_errors():
    if len(missing_rep["CS"]) > 0:
        for e in missing_rep["CS"]:
            errors.append({'Code': 'MS07', 'ErrorLevel': 'CRITICAL',
                           'ObjectID': f'{e}',
                           'ObjectType': 'Concept',
                           'Message': f'Missing Concept Scheme {e}'}
                          )
        missing_rep["CS"].clear()
    if len(missing_rep["CL"]) > 0:
        for e in missing_rep["CL"]:
            errors.append({'Code': 'MS02', 'ErrorLevel': 'CRITICAL',
                           'ObjectID': f'{e}',
                           'ObjectType': 'Codelist',
                           'Message': f'Missing Codelist {e}'})
        missing_rep["CL"].clear()
    if len(missing_rep["CON"]) > 0:
        for e in missing_rep["CON"]:
            errors.append({'Code': 'MS03', 'ErrorLevel': 'CRITICAL',
                           'ObjectID': f'{e}',
                           'ObjectType': 'Concept',
                           'Message': f'Missing Concept {e}'})
        missing_rep["CON"].clear()


def create_metadata(json_meta):
    """
        Metadata validations stands for the next schema:

        .. list-table:: Metadata validations
            :widths: 20 80
            :header-rows: 1

            * - Code
              - Description
            * - MS01
              - Check that the metadata file contains at least one DSD
            * - MS02
              - Check that the metadata file contains related codelists
            * - MS03
              - Check if the DSD metadata file contains the concepts needed \
                for each DSD
            * - MS04
              - Check if the dimensions in Attribute Relationship are in \
                DimensionList in DSD file
            * - MS05
              - Check if the primary measure in Attribute Relationship is in
                MeasureList in DSD file
            * - MS06
              - Check if all DSDs present in the metadata file are unique
            * - MS07
              - Check if all Concept Scheme needed are present
            * - MX01
              - Check if minimum structural requirements for DSD xml are \
                satisfied
            * - MX02
              - Check if every DSD has primary measure defined
    """
    # Reset dict to store metadata
    metadata = dict()

    if ORGS in json_meta:
        metadata[ORGS] = create_organisations(json_meta[ORGS])
    if CODELISTS in json_meta:
        metadata[CODELISTS] = create_scheme(json_meta[CODELISTS], CL, CODE)
        codelists.update(metadata[CODELISTS])
    if CONCEPTS in json_meta:
        metadata[CONCEPTS] = create_scheme(json_meta[CONCEPTS], CS, CON)
        concepts.update(metadata[CONCEPTS])
    if DSDS in json_meta:
        metadata[DSDS] = create_datastructures(json_meta[DSDS])
        datastructures.update(metadata[DSDS])
    if DATAFLOWS in json_meta:
        metadata[DATAFLOWS] = create_dataflows(json_meta[DATAFLOWS])
        dataflows.update(metadata[DATAFLOWS])
    if CONSTRAINTS in json_meta:
        metadata[CONSTRAINTS] = create_constraints(json_meta[CONSTRAINTS])

    grouping_errors()

    metadata['errors'] = copy.copy(errors)

    # Reset global variables
    agencies.clear()
    concepts.clear()
    codelists.clear()
    datastructures.clear()
    dataflows.clear()
    errors.clear()
    return metadata



from datetime import datetime
from io import StringIO

import pandas as pd

from sdmxthon.model.component import PrimaryMeasure
from sdmxthon.model.header import Header, Sender, Party
from sdmxthon.parsers.data_validations import get_mandatory_attributes
from sdmxthon.utils.enums import MessageTypeEnum
from sdmxthon.utils.handlers import add_indent
from sdmxthon.utils.mappings import messageAbbr, commonAbbr, genericAbbr, \
    structureSpecificAbbr, structureAbbr
from sdmxthon.utils.parsing_words import ORGS, DATAFLOWS, CODELISTS, \
    CONCEPTS, DSDS, CONSTRAINTS

chunksize = 100000


#
#      --------------------------------------------
#     |                                            |
#     |                   Common                   |
#     |                                            |
#      --------------------------------------------
#


def addStructure(dataset, prettyprint, type_):
    outfile = ''

    if prettyprint:
        child2 = '\t\t'
        child3 = '\t\t\t'
        child4 = '\t\t\t\t'
        nl = '\n'
    else:
        child2 = child3 = child4 = nl = ''

    outfile += f'{nl}{child2}<{messageAbbr}:Structure ' \
               f'structureID="{dataset.structure.id}" '
    if type_ != MessageTypeEnum.GenericDataSet:
        outfile += f'namespace="urn:sdmx:org.sdmx.infomodel.' \
                   f'datastructure.DataStructure=' \
                   f'{dataset.structure.agencyID}:{dataset.structure.id}' \
                   f'({dataset.structure.version})" '

    outfile += f'dimensionAtObservation="{dataset.dim_at_obs}">{nl}'
    outfile += f'{child3}<{commonAbbr}:Structure>{nl}{child4}' \
               f'<Ref agencyID="{dataset.structure.agencyID}" ' \
               f'id="{dataset.structure.id}" ' \
               f'version="{dataset.structure.version}" ' \
               f'class="DataStructure"/>' \
               f'{nl}{child3}</{commonAbbr}:Structure>' \
               f'{nl}{child2}</{messageAbbr}:Structure>'

    return outfile


def create_namespaces(data_type, payload, type_, prettyprint):
    if prettyprint:
        nl = '\n'
    else:
        nl = ''
    outfile = f'<?xml version="1.0" encoding="UTF-8"?>{nl}'
    outfile += f'<{messageAbbr}:{data_type} ' \
               f'xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance" ' \
               f'xmlns:{messageAbbr}="http://www.sdmx.org/resources/sdmxml' \
               f'/schemas/v2_1/message" '
    if type_ == MessageTypeEnum.GenericDataSet:
        outfile += f'xmlns:{genericAbbr}="http://www.sdmx.org/resources' \
                   f'/sdmxml/schemas/v2_1/data/generic" '
    elif type_ == MessageTypeEnum.StructureDataSet:
        outfile += f'xmlns:{structureSpecificAbbr}="http://www.sdmx.org' \
                   f'/resources/sdmxml/schemas/v2_1/data/structurespecific" '
        if isinstance(payload, dict):
            count = 0
            for key, record in payload.items():
                count += 1
                if record.structure is None:
                    raise Exception(f'Dataset {key} has no structure defined')
                outfile += f'xmlns:ns{count}="urn:sdmx:org.sdmx.infomodel' \
                           f'.datastructure.DataStructure=' \
                           f'{record.structure.agencyID}:' \
                           f'{record.structure.id}' \
                           f'({record.structure.version})' \
                           f':ObsLevelDim:{record.dim_at_obs}" '
        else:
            if payload.structure is None:
                raise Exception('Dataset has no structure defined')
            outfile += f'xmlns:ns1="urn:sdmx:org.sdmx.infomodel' \
                       f'.datastructure.DataStructure=' \
                       f'{payload.structure.agencyID}:{payload.structure.id}' \
                       f'({payload.structure.version})' \
                       f':ObsLevelDim:{payload.dim_at_obs}" '
    else:
        outfile += f'xmlns:{structureAbbr}="http://www.sdmx.org/resources' \
                   f'/sdmxml/schemas/v2_1/structure" '
    outfile += f'xmlns:{commonAbbr}="http://www.sdmx.org/resources/sdmxml' \
               f'/schemas/v2_1/common" ' \
               f'xsi:schemaLocation="http://www.sdmx.org/resources/sdmxml' \
               f'/schemas/v2_1/message ' \
               f'https://registry.sdmx.org/schemas/v2_1/SDMXMessage.xsd">{nl}'

    return outfile


def write_from_header(header, prettyprint):
    if prettyprint:
        child1 = '\t'
        child2 = '\t\t'
        nl = '\n'
    else:
        child1 = child2 = nl = ''

    outfile = f'{child1}<{messageAbbr}:Header>{nl}'
    if header.id_ is not None:
        outfile += f'{child2}<{messageAbbr}:ID>{header.id_}</{messageAbbr}:ID>'
    else:
        outfile += f'{child2}<{messageAbbr}:ID>test</{messageAbbr}:ID>'
    outfile += f'{nl}{child2}<{messageAbbr}:Test>'
    if header.test is not None:
        outfile += f'{str(header.test).lower()}'
    else:
        outfile += 'true'
    outfile += f'</{messageAbbr}:Test>'
    outfile += f'{nl}{child2}<{messageAbbr}:Prepared>'
    if header.prepared is not None:
        outfile += f'{header.prepared.strftime("%Y-%m-%dT%H:%M:%S")}'
    else:
        outfile += f'{datetime.now().strftime("%Y-%m-%dT%H:%M:%S")}'
    outfile += f'</{messageAbbr}:Prepared>'

    outfile += f'{nl}{child2}<{messageAbbr}:Sender '
    if header.sender is not None:
        outfile += f'id="{header.sender.id_}"/>'
    else:
        outfile += 'id="Unknown"/>'

    if header.receiver is not None and len(header.receiver) > 0:
        for receiver in header.receiver:
            outfile += f'{nl}{child2}<{messageAbbr}:Receiver '
            outfile += f'id="{receiver.id_}"/>'
    else:
        outfile += f'{nl}{child2}<{messageAbbr}:Receiver '
        outfile += 'id="Not_supplied"/>'

    return outfile


def process_dataset(dataset):
    dataset.data = dataset.data.fillna(value='')
    return dataset


def _write_metadata_element(payload, dict_word, pkg_word,
                            item_word, newline, indent):
    outfile = ""
    if dict_word in payload:
        indent_child = newline + add_indent(indent)
        outfile += f'{indent_child}<{structureAbbr}:{pkg_word}>'
        for e in payload[dict_word].values():
            outfile += e._parse_XML(indent_child,
                                    f'{structureAbbr}:{item_word}')
        outfile += f'{indent_child}</{structureAbbr}:{pkg_word}>'

    return outfile


def parse_metadata(payload, prettyprint):
    if prettyprint:
        indent = '\t'
        newline = '\n'
    else:
        indent = newline = ''

    outfile = f'{indent}<{messageAbbr}:Structures>'
    outfile += _write_metadata_element(payload, ORGS, 'OrganisationSchemes',
                                       'AgencyScheme', newline, indent)
    outfile += _write_metadata_element(payload, DATAFLOWS, 'Dataflows',
                                       'Dataflow', newline, indent)
    outfile += _write_metadata_element(payload, CODELISTS, 'Codelists',
                                       'Codelist', newline, indent)
    outfile += _write_metadata_element(payload, CONCEPTS, 'Concepts',
                                       'ConceptScheme', newline, indent)
    outfile += _write_metadata_element(payload, DSDS, 'DataStructures',
                                       'DataStructure', newline, indent)
    outfile += _write_metadata_element(payload, CONSTRAINTS, 'Constraints',
                                       'ContentConstraint', newline, indent)

    outfile += f'{newline}{indent}</{messageAbbr}:Structures>{newline}'

    return outfile


def writer(path, payload, type_, prettyprint=True, id_='test',
           test='true',
           prepared=datetime.now(),
           sender='Unknown',
           receiver='Not_supplied',
           header: Header = None):
    if prettyprint:
        child1 = '\t'
        child2 = '\t\t'
        nl = '\n'
    else:
        child1 = child2 = nl = ''

    if type_ == MessageTypeEnum.GenericDataSet:
        data_type_string = 'GenericData'
    elif type_ == MessageTypeEnum.StructureDataSet:
        data_type_string = 'StructureSpecificData'
    else:
        data_type_string = 'Structure'

    # Header
    outfile = create_namespaces(data_type_string, payload, type_, prettyprint)

    if header is None:
        header = Header(id_, test, prepared, Sender(sender), [Party(receiver)])

    outfile += write_from_header(header, prettyprint)

    if isinstance(payload, dict) and type_ is not MessageTypeEnum.Metadata:
        for record in payload.values():
            outfile += addStructure(record, prettyprint, type_)
    elif type_ is not MessageTypeEnum.Metadata:
        outfile += addStructure(payload, prettyprint, type_)

    if type_ is not MessageTypeEnum.Metadata and header is not None:
        if header.dataset_action is not None:
            outfile += f'{nl}{child2}<{messageAbbr}:DataSetAction>' \
                       f'{header.dataset_action}</{messageAbbr}:DataSetAction>'
    if header is not None:
        if header.source is not None:
            list_names = header.source._to_XML(name=f'{messageAbbr}:Source',
                                               prettyprint=True)
            for elem in list_names:
                outfile += f'{nl}{child1}' + elem
        else:
            outfile += f'{nl}{child2}<{messageAbbr}:Source xml:lang="en">' \
                       f'SDMXthon</{messageAbbr}:Source>'

    outfile += f'{nl}{child1}</{messageAbbr}:Header>{nl}'

    # Dataset
    if type_ == MessageTypeEnum.GenericDataSet:
        if isinstance(payload, dict):
            for record in payload.values():
                record = process_dataset(record)
                outfile += genWriting(record, prettyprint, record.dim_at_obs)
        else:
            payload = process_dataset(payload)
            outfile += genWriting(payload, prettyprint, dim=payload.dim_at_obs)
    elif type_ == MessageTypeEnum.StructureDataSet:
        if isinstance(payload, dict):
            count = 0
            for record in payload.values():
                count += 1
                record = process_dataset(record)
                outfile += strWriting(record, prettyprint, count,
                                      record.dim_at_obs)
        else:
            payload = process_dataset(payload)
            outfile += strWriting(payload, prettyprint, dim=payload.dim_at_obs)
    elif type_ == MessageTypeEnum.Metadata:
        if len(payload) > 0:
            outfile += parse_metadata(payload, prettyprint)
    outfile += f'</{messageAbbr}:{data_type_string}>'

    if path != '':
        with open(path, "w", encoding="UTF-8", errors='replace') as f:
            f.write(outfile)
    else:
        f = StringIO()
        f.write(outfile)
        return f


def format_dict_ser(out, parser, data_dict, obs):
    data_dict['Series'][0]['Obs'] = obs.to_dict(orient="records")
    out.append(parser(data_dict['Series'][0]))
    del data_dict['Series'][0]


def series_process(parser, data, data_dict, series_codes, obs_codes):
    out = []
    if all(elem in data.columns for elem in obs_codes):
        data.groupby(by=series_codes)[obs_codes].apply(
            lambda x: format_dict_ser(out, parser, data_dict, x))

    return ''.join(out)


def get_codes(dim, dataset):
    series_codes = []
    obs_codes = [dim, dataset.structure.measure_code]
    for e in dataset.structure.attribute_descriptor.components.values():
        if e.id in dataset.data.keys() and isinstance(e.related_to,
                                                      PrimaryMeasure):
            obs_codes.append(e.id)
    for e in dataset.data.keys():
        if ((e in dataset.structure.dimension_codes and e != dim)
                or (e in dataset.structure.attribute_codes and
                    e not in obs_codes)):
            series_codes.append(e)

    return series_codes, obs_codes


#
#      --------------------------------------------
#     |                                            |
#     |              Structure Specific            |
#     |                                            |
#      --------------------------------------------
#

def memory_optimization_str(dataset, opt_att_codes, prettyprint):
    outfile = ""
    length_ = len(dataset.data)
    if len(dataset.data) > chunksize:
        previous = 0
        next_ = chunksize
        while previous <= length_:
            outfile += obs_str(dataset.data.iloc[previous:next_],
                               opt_att_codes, prettyprint)
            previous = next_
            next_ += chunksize

            if next_ >= length_:
                outfile += obs_str(dataset.data.iloc[previous:],
                                   opt_att_codes, prettyprint)
                previous = next_
    else:
        outfile += obs_str(dataset.data, opt_att_codes, prettyprint)

    return outfile


def strWriting(dataset, prettyprint=True, count=1, dim="AllDimensions"):
    outfile = ''

    if prettyprint:
        child1 = '\t'
        nl = '\n'
    else:
        child1 = nl = ''

    attached_attributes_str = ''
    for k, v in dataset.attached_attributes.items():
        attached_attributes_str += f'{k}="{v}" '

    # Datasets
    outfile += f'{child1}<{messageAbbr}:DataSet {attached_attributes_str}' \
               f'ss:structureRef="{dataset.structure.id}" ' \
               f'xsi:type="ns{count}:DataSetType" ' \
               f'ss:dataScope="DataStructure" ' \
               f'action="Replace">{nl}'
    man_att = get_mandatory_attributes(dataset.structure)
    opt_att_codes = [att for att in dataset.structure.attribute_codes
                     if att not in man_att]

    if dim == "AllDimensions":
        outfile += memory_optimization_str(dataset, opt_att_codes, prettyprint)
    else:
        series_codes, obs_codes = get_codes(dim, dataset)
        outfile += ser_str(dataset.data, opt_att_codes, series_codes,
                           obs_codes, prettyprint)

    outfile += f'{child1}</{messageAbbr}:DataSet>{nl}'

    return outfile


def format_obs_str(data: dict, prettyprint: bool) -> str:
    if prettyprint:
        child2 = '\t\t'
        nl = '\n'
    else:
        child2 = nl = ''

    out = f"{child2}<Obs "

    for k, v in data.items():
        out += f'{k}="{v}" '

    out += f"/>{nl}"

    return out


def obs_str(data: pd.DataFrame,
            codes: list,
            prettyprint=True) -> str:
    parser = lambda x: format_obs_str(x, prettyprint)  # noqa: E731

    iterator = map(parser, data.to_dict(orient='records'))
    out = ''.join(iterator)

    for c in codes:
        out = out.replace(f'{c}="" ', '')

    return out


def format_ser_str(data: dict, prettyprint: bool) -> str:
    if prettyprint:
        child2 = '\t\t'
        child3 = '\t\t\t'
        nl = '\n'
    else:
        child2 = child3 = nl = ''

    out = f"{child2}<Series "

    for k, v in data.items():
        if k != 'Obs':
            out += f'{k}="{v}" '

    out += f">{nl}"

    for obs in data['Obs']:
        out += f"{child3}<Obs "

        for k, v in obs.items():
            out += f'{k}="{v}" '

        out += f"/>{nl}"

    out += f"{child2}</Series>{nl}"

    return out


def ser_str(data: pd.DataFrame,
            opt_att_codes: list,
            series_codes: list,
            obs_codes: list,
            prettyprint=True) -> str:
    # Getting each datapoint from data and creating dict
    data = data.sort_values(series_codes, axis=0)
    data_dict = {'Series': data[series_codes].drop_duplicates().reset_index(
        drop=True).to_dict(orient="records")}

    parser = lambda x: format_ser_str(data=x,  # noqa: E731
                                      prettyprint=prettyprint)

    out = series_process(parser=parser, data=data, data_dict=data_dict,
                         series_codes=series_codes, obs_codes=obs_codes)

    for c in opt_att_codes:
        out = out.replace(f'{c}="" ', '')

    return out


#
#      --------------------------------------------
#     |                                            |
#     |                   Generic                  |
#     |                                            |
#      --------------------------------------------
#

def memory_optimization_gen(dataset, dim_codes, att_codes, measure_code,
                            prettyprint):
    outfile = ""
    length_ = len(dataset.data)
    if len(dataset.data) > chunksize:
        previous = 0
        next_ = chunksize
        while previous <= length_:
            outfile += obs_gen(dataset.data.iloc[previous:next_],
                               dim_codes=dim_codes,
                               att_codes=att_codes,
                               measure_code=measure_code,
                               prettyprint=prettyprint)
            previous = next_
            next_ += chunksize

            if next_ >= length_:
                outfile += obs_gen(dataset.data.iloc[previous:length_],
                                   dim_codes=dim_codes,
                                   att_codes=att_codes,
                                   measure_code=measure_code,
                                   prettyprint=prettyprint)
                previous = next_
    else:
        outfile += obs_gen(dataset.data,
                           dim_codes=dim_codes,
                           att_codes=att_codes,
                           measure_code=measure_code,
                           prettyprint=prettyprint)

    return outfile


def genWriting(dataset, prettyprint=True, dim="AllDimensions"):
    outfile = ''

    if prettyprint:
        child1 = '\t'
        child2 = '\t\t'
        child3 = '\t\t\t'
        nl = '\n'
    else:
        child1 = child2 = child3 = nl = ''

    outfile += f'{child1}<{messageAbbr}:DataSet ' \
               f'structureRef="{dataset.structure.id}" action="Replace">{nl}'
    if len(dataset.attached_attributes) > 0:
        outfile += f'{child2}<{genericAbbr}:Attributes>{nl}'
        for k, v in dataset.attached_attributes.items():
            outfile += f'{child3}<{genericAbbr}:Value id="{k}" ' \
                       f'value="{v}"/>{nl}'
        outfile += f'{child2}</{genericAbbr}:Attributes>{nl}'

    dim_codes = [v for v in dataset.structure.dimension_codes if
                 v in dataset.data.columns]
    att_codes = [v for v in dataset.structure.attribute_codes if
                 v in dataset.data.columns]
    measure_code = dataset.structure.measure_code

    if dim == "AllDimensions":
        outfile += memory_optimization_gen(dataset, dim_codes, att_codes,
                                           measure_code, prettyprint)
    else:
        series_codes, obs_codes = get_codes(dim, dataset)

        outfile += ser_gen(dataset.data,
                           dim_codes=dim_codes,
                           att_codes=att_codes,
                           measure_code=measure_code,
                           prettyprint=prettyprint,
                           series_codes=series_codes,
                           obs_codes=obs_codes)

    outfile += f'{child1}</{messageAbbr}:DataSet>{nl}'

    return outfile


def format_measure_att(data, measure_code, att_codes, child3, child4, nl):
    out = ""
    if measure_code != "OBS_VALUE":
        out += f'{child3}<{genericAbbr}:ObsValue id="{measure_code}" ' \
               f'value="{data[measure_code]}"/>{nl}'
    else:
        out += f'{child3}<{genericAbbr}:ObsValue ' \
               f'value="{data[measure_code]}"/>{nl}'

    if len(att_codes) > 0:
        out += f"{child3}<{genericAbbr}:Attributes>{nl}"

        for k in att_codes:
            if data[k] != '':
                out += f'{child4}<{genericAbbr}:Value id="{k}" ' \
                       f'value="{data[k]}"/>{nl}'

        out += f"{child3}</{genericAbbr}:Attributes>{nl}"

    return out


def format_obs(data: dict,
               dim_codes: list,
               att_codes: list,
               measure_code: str,
               prettyprint=True):
    if prettyprint:
        child2 = '\t\t'
        child3 = '\t\t\t'
        child4 = '\t\t\t\t'
        nl = '\n'
    else:
        child2 = child3 = child4 = nl = ''

    out = f"{child2}<{genericAbbr}:Obs>{nl}"

    out += f"{child3}<{genericAbbr}:ObsKey>{nl}"

    for k in dim_codes:
        out += f'{child4}<{genericAbbr}:Value id="{k}" value="{data[k]}"/>{nl}'

    out += f"{child3}</{genericAbbr}:ObsKey>{nl}"

    out += format_measure_att(data, measure_code, att_codes,
                              child3, child4, nl)

    out += f"{child2}</{genericAbbr}:Obs>{nl}"

    return out


def obs_gen(data: pd.DataFrame,
            dim_codes: list,
            att_codes: list,
            measure_code: str,
            prettyprint=True):
    parser = lambda x: format_obs(x, dim_codes, att_codes,  # noqa: E731
                                  measure_code, prettyprint)

    iterator = map(parser, data.to_dict(orient='records'))
    out = ''.join(iterator)

    return out


def format_ser(data: dict,
               measure_code: str,
               series_key: list,
               series_attr: list,
               obs_attr: list,
               dim: str,
               prettyprint=True):
    if prettyprint:
        child2 = '\t\t'
        child3 = '\t\t\t'
        child4 = '\t\t\t\t'
        nl = '\n'
    else:
        child2 = child3 = child4 = nl = ''

    out = f"{child2}<{genericAbbr}:Series>{nl}"
    # --------------  Series --------------
    # Series Keys

    out += f"{child3}<{genericAbbr}:SeriesKey>{nl}"

    for k in series_key:
        out += f'{child4}<{genericAbbr}:Value id="{k}" value="{data[k]}"/>{nl}'

    out += f"{child3}</{genericAbbr}:SeriesKey>{nl}"

    # Series Attributes

    if len(series_attr) > 0:
        out += f"{child3}<{genericAbbr}:Attributes>{nl}"

        for k in series_attr:
            if data[k] != '':
                out += f'{child4}<{genericAbbr}:Value id="{k}" ' \
                       f'value="{data[k]}"/>{nl}'

        out += f"{child3}</{genericAbbr}:Attributes>{nl}"

    # --------------  Obs  --------------

    for obs in data['Obs']:
        out += f"{child3}<{genericAbbr}:Obs>{nl}"
        # Obs Elements
        if dim in obs:
            out += f'{child4}<{genericAbbr}:ObsDimension ' \
                   f'value="{obs[dim]}"/>{nl}'

        out += format_measure_att(obs, measure_code, obs_attr,
                                  child3, child4, nl)

        out += f"{child3}</{genericAbbr}:Obs>{nl}"

    out += f"{child2}</{genericAbbr}:Series>{nl}"

    del data['Obs']

    return out


def ser_gen(data: pd.DataFrame,
            dim_codes: list,
            att_codes: list,
            measure_code: str,
            obs_codes: list,
            series_codes: list,
            prettyprint=True):
    # Getting each datapoint from data and creating dict

    series_key = [v for v in series_codes if v in dim_codes]
    series_att = [v for v in series_codes if v in att_codes]
    dim = obs_codes[0]
    if len(obs_codes) == 2:
        obs_att = []
    else:
        obs_att = obs_codes[2:]

    data = data.sort_values(series_codes, axis=0)

    data_dict = {'Series': data[series_codes].drop_duplicates().reset_index(
        drop=True).to_dict(orient="records")}

    parser = lambda x: format_ser(data=x,  # noqa: E731
                                  measure_code=measure_code,
                                  series_key=series_key,
                                  series_attr=series_att,
                                  obs_attr=obs_att,
                                  dim=dim,
                                  prettyprint=prettyprint)

    out = series_process(parser=parser, data=data, data_dict=data_dict,
                         series_codes=series_codes, obs_codes=obs_codes)

    del data_dict

    return out



from sdmxthon.model.base import IdentifiableArtefact
from sdmxthon.model.component import Dimension, Attribute, PrimaryMeasure
from sdmxthon.utils.handlers import add_indent, export_intern_data


class GroupDimension(object):
    """ Parser of a Dimension in a GroupDimensionDescriptor"""

    def __init__(self):
        self._ref = None

    def __eq__(self, other):
        if isinstance(other, GroupDimension):
            return super(GroupDimension, self).__eq__(other)

        return False

    @property
    def ref(self):
        """Reference to the dimension"""
        return self._ref


class ComponentList(IdentifiableArtefact):
    """An abstract definition of a list of components.
       A concrete example is a Dimension Descriptor which defines
       the list of Dimensions in a DataStructure Definition.
    """
    _componentType = None

    def __init__(self, id_: str = None, uri: str = None, urn: str = None,
                 annotations=None, components=None):

        super(ComponentList, self).__init__(id_=id_, uri=uri, urn=urn,
                                            annotations=annotations)
        self._components = {}
        if components is not None:
            for c in components.values():
                self.add_component(c)

    def __eq__(self, other):
        if isinstance(other, ComponentList):
            return (super(ComponentList, self).__eq__(other) and
                    self._components == other._components)

        return False

    def __str__(self):
        return f'<{self.__class__.__name__} - {self.id}>'

    def __unicode__(self):
        return f'<{self.__class__.__name__} - {self.id}>'

    def __repr__(self):
        return f'<{self.__class__.__name__} - {self.id}>'

    @property
    def components(self):
        """An aggregate association to one or more components
        which make up the list."""
        return self._components

    def add_component(self, value):
        """Method to add a Component to the ComponentList"""
        if isinstance(self, MeasureDescriptor) and len(self._components) == 1:
            raise ValueError('Measure Descriptor cannot have '
                             'more than one Primary Measure')
        if isinstance(value, (Dimension, Attribute, PrimaryMeasure)):
            value.componentList = self
            self._components[value.id] = value
        elif isinstance(value, GroupDimension):
            value.componentList = self
            self._components[value.ref] = value
        else:
            raise TypeError(
                f"The object has to be of the dim_type "
                f"[Dimension, Attribute, PrimaryMeasure], "
                f"{value.__class__.__name__} provided")

    def __len__(self):
        return len(self.components)

    def __getitem__(self, value):
        return self.components[value]

    def _parse_XML(self, indent, label):
        prettyprint = indent != ''

        indent = add_indent(indent)

        data = super(ComponentList, self)._to_XML(prettyprint)

        outfile = ''

        attributes = data.get('Attributes') or None

        if attributes is not None:
            outfile += f'{indent}<{label}{attributes}>'
        else:
            outfile += f'{indent}<{label}>'

        outfile += export_intern_data(data, indent)

        for i in self.components.values():
            outfile += i._parse_XML(indent, self._componentType)

        outfile += f'{indent}</{label}>'

        return outfile


class DimensionDescriptor(ComponentList, object):
    """An ordered set of metadata concepts that, combined, classify a
    statistical series, and whose values, when combined (the key) in an
    instance such as a data set, uniquely identify a specific observation
    """
    _componentType = "Dimension"

    def __init__(self, id_: str = None, uri: str = None, urn: str = None,
                 annotations=None, components=None):

        if components is None:
            components = {}
        super(DimensionDescriptor, self).__init__(id_=id_, uri=uri, urn=urn,
                                                  annotations=annotations,
                                                  components=components)

    def __eq__(self, other):
        if isinstance(other, DimensionDescriptor):
            return super(DimensionDescriptor, self).__eq__(other)

        return False


class AttributeDescriptor(ComponentList, object):
    """A set metadata concepts that define the attributes of a
    Data Structure Definition."""

    _componentType = "Attribute"

    def __init__(self, id_: str = None, uri: str = None, urn: str = None,
                 annotations=None, components=None):
        if components is None:
            components = {}

        super(AttributeDescriptor, self).__init__(id_=id_, uri=uri, urn=urn,
                                                  annotations=annotations,
                                                  components=components)

    def __eq__(self, other):
        if isinstance(other, AttributeDescriptor):
            return super(AttributeDescriptor, self).__eq__(other)

        return False


class MeasureDescriptor(ComponentList, object):
    """A metadata concept that defines the measure of a
    Data Structure Definition"""

    _componentType = "PrimaryMeasure"

    def __init__(self, id_: str = None, uri: str = None, urn: str = None,
                 annotations=None, components=None):

        if components is None:
            components = {}
        super(MeasureDescriptor, self).__init__(id_=id_, uri=uri, urn=urn,
                                                annotations=annotations,
                                                components=components)

    def __eq__(self, other):
        if isinstance(other, MeasureDescriptor):
            return super(MeasureDescriptor, self).__eq__(other)

        return False


class GroupDimensionDescriptor(ComponentList, object):
    """A set metadata concepts that define a partial key derived from the
    Dimension Descriptor in a Data Structure Definition.
    """

    _componentType = "Dimension"

    def __init__(self, id_: str = None, uri: str = None, urn: str = None,
                 annotations=None, components=None):
        if components is None:
            components = {}
        super(GroupDimensionDescriptor, self). \
            __init__(id_=id_, uri=uri, urn=urn,
                     annotations=annotations, components=components)

    def __eq__(self, other):
        if isinstance(other, GroupDimensionDescriptor):
            return super(GroupDimensionDescriptor, self).__eq__(other)

        return False



"""itemScheme file contains the classes for Item and ItemScheme and its
derivatives """

import warnings
from datetime import datetime
from typing import List

from sdmxthon.model.base import MaintainableArtefact, NameableArtefact, \
    InternationalString
from sdmxthon.model.header import Contact
from sdmxthon.model.representation import Representation
from sdmxthon.model.utils import generic_setter, bool_setter
from sdmxthon.utils.handlers import add_indent, export_intern_data, \
    split_unique_id
from sdmxthon.utils.mappings import structureAbbr


class ItemScheme(MaintainableArtefact):
    """The descriptive information for an arrangement or division of objects
    into groups based on characteristics, which the objects have in common.
    """

    _itemType = None

    def __init__(self, id_: str = None, uri: str = None, urn: str = None,
                 annotations=None,
                 name: str = None, description: str = None,
                 version: str = None, validFrom: datetime = None,
                 validTo: datetime = None,
                 isFinal: bool = None, isExternalReference: bool = None,
                 serviceUrl: str = None,
                 structureUrl: str = None, maintainer=None,
                 items=None, isPartial: bool = None):

        super(ItemScheme, self). \
            __init__(id_=id_, uri=uri, urn=urn,
                     annotations=annotations,
                     name=name, description=description,
                     version=version, validFrom=validFrom,
                     validTo=validTo,
                     isFinal=isFinal,
                     isExternalReference=isExternalReference,
                     serviceUrl=serviceUrl,
                     structureUrl=structureUrl,
                     maintainer=maintainer)

        self._isPartial = isPartial

        self._items = {}
        if items is not None:
            urn_list = []
            for i in items:
                if i.urn is not None and i.urn not in urn_list:
                    urn_list.append(i.urn)
                    self.append(i)
                elif i.urn is not None:
                    raise ValueError(
                        'Item Scheme cannot have two items with same URN')

    def __eq__(self, other):
        if isinstance(other, ItemScheme):
            return super(ItemScheme, self).__eq__(
                other) and self._items == other._items
        return False

    def __str__(self):
        return f'<{self.__class__.__name__} - {self.unique_id}>'

    def __unicode__(self):
        return f'<{self.__class__.__name__} - {self.unique_id}>'

    def __repr__(self):
        return f'<{self.__class__.__name__} - {self.unique_id}>'

    @property
    def items(self):
        """Association to the Items in the scheme"""
        return self._items

    @property
    def is_partial(self):
        """Denotes whether the Item Scheme contains a sub set of the
        full set of Items in the maintained scheme."""
        return self._isPartial

    @is_partial.setter
    def is_partial(self, value):
        self._isPartial = bool_setter(value)

    def append(self, value):

        """Adds an Item to the ItemScheme"""

        if isinstance(value, globals()[self._itemType]):
            if value.id is None:
                warnings.warn("Item not added because it did not have id_")
            if value.id not in self._items.keys():
                self._items[value.id] = value
                value.scheme = self
                if (value.parent is not None and
                        value.parent in self._items.keys()):
                    self._items[value.parent].add_child(value.id)
                    value.parent = self._items[value.parent]
        else:
            raise TypeError(
                f"The object has to be of the type {self._itemType}")

    def _parse_XML(self, indent, label):
        prettyprint = indent != ''

        indent = add_indent(indent)

        data = super(ItemScheme, self)._to_XML(prettyprint)

        if self.is_partial is not None:
            data['Attributes'] += f' isPartial="' \
                                  f'{str(self.is_partial).lower()}"'

        outfile = ''

        attributes = data.get('Attributes') or None

        if attributes is not None:
            outfile += f'{indent}<{label}{attributes}>'
        else:
            outfile += f'{indent}<{label}>'

        outfile += export_intern_data(data, indent)

        for i in self.items.values():
            outfile += i._parse_XML(indent, self._itemType)

        outfile += f'{indent}</{label}>'

        return outfile


class ConceptScheme(ItemScheme):
    """ The descriptive information for an arrangement or division of
    concepts into groups based on characteristics, which the objects have in
    common.

        ItemType:
            :obj:`Concept`

    """

    _itemType = "Concept"

    def __init__(self, id_: str = None, uri: str = None, urn: str = None,
                 annotations=None,
                 name: InternationalString = None,
                 description: InternationalString = None,
                 version: str = None, validFrom: datetime = None,
                 validTo: datetime = None,
                 isFinal: bool = None, isExternalReference: bool = None,
                 isPartial: bool = None,
                 serviceUrl: str = None,
                 structureUrl: str = None, maintainer=None,
                 items=None):
        if annotations is None:
            annotations = []
        self._cl_references = {}
        super(ConceptScheme, self). \
            __init__(id_=id_, uri=uri, urn=urn,
                     annotations=annotations,
                     name=name, description=description,
                     version=version,
                     validFrom=validFrom,
                     validTo=validTo,
                     isFinal=isFinal,
                     isExternalReference=isExternalReference,
                     isPartial=isPartial,
                     serviceUrl=serviceUrl,
                     structureUrl=structureUrl,
                     maintainer=maintainer,
                     items=items)

        self._checked = False

    def __eq__(self, other):
        if isinstance(other, ConceptScheme):
            if not self._checked:
                self._checked = True
                return super(ConceptScheme, self).__eq__(other)

            return True

        return False

    @property
    def cl_references(self):
        """For parsing purposes, makes a dict of all the Codelist references"""
        return self._cl_references


class Codelist(ItemScheme):
    """ A list from which some statistical concepts (coded concepts) take
    their values.

        ItemType:
            :obj:`Code`
    """

    _itemType = "Code"

    def __init__(self, id_: str = None, uri: str = None, urn: str = None,
                 annotations=None,
                 name: str = None, description: str = None,
                 version: str = None, validFrom: datetime = None,
                 validTo: datetime = None,
                 isFinal: bool = None, isExternalReference: bool = None,
                 isPartial: bool = None,
                 serviceUrl: str = None,
                 structureUrl: str = None, maintainer=None,
                 items=None):
        super(Codelist, self).__init__(id_=id_, uri=uri, urn=urn,
                                       annotations=annotations,
                                       name=name, description=description,
                                       version=version, validFrom=validFrom,
                                       validTo=validTo,
                                       isFinal=isFinal,
                                       isExternalReference=isExternalReference,
                                       serviceUrl=serviceUrl,
                                       structureUrl=structureUrl,
                                       maintainer=maintainer,
                                       isPartial=isPartial,
                                       items=items)
        self._checked = False

    def __eq__(self, other):
        if isinstance(other, Codelist):
            if not self._checked:
                self._checked = True
                return super(Codelist, self).__eq__(other)

            return True

        return False

    def __str__(self):
        return f'<{self.__class__.__name__} - {self.id}>'

    def __unicode__(self):
        return f'<{self.__class__.__name__} - {self.id}>'

    def __repr__(self):
        return f'<{self.__class__.__name__} - {self.id}>'


class OrganisationScheme(ItemScheme):
    """Abstract class. Used for structure messages"""

    def __init__(self, id_: str = None, uri: str = None, urn: str = None,
                 annotations=None,
                 name: InternationalString = None,
                 description: InternationalString = None,
                 version: str = None, validFrom: datetime = None,
                 validTo: datetime = None,
                 isFinal: bool = None,
                 isExternalReference: bool = None,
                 isPartial: bool = None,
                 serviceUrl: str = None,
                 structureUrl: str = None,
                 maintainer=None,
                 items=None):

        super(OrganisationScheme, self). \
            __init__(id_=id_, uri=uri, urn=urn,
                     annotations=annotations,
                     name=name,
                     description=description,
                     version=version,
                     validFrom=validFrom,
                     validTo=validTo,
                     isFinal=isFinal,
                     isExternalReference=isExternalReference,
                     isPartial=isPartial,
                     serviceUrl=serviceUrl,
                     structureUrl=structureUrl,
                     maintainer=maintainer,
                     items=items)

    def __eq__(self, other):
        if isinstance(other, OrganisationScheme):
            return super(OrganisationScheme, self).__eq__(other)

        return False


class AgencyScheme(OrganisationScheme):
    """ A maintained collection of Maintenance Agencies.

        ItemType:
            :obj:`Agency`
    """

    _itemType = "Agency"

    def __init__(self, id_: str = None, uri: str = None, urn: str = None,
                 annotations=None,
                 name: InternationalString = None,
                 description: InternationalString = None,
                 version: str = None, validFrom: datetime = None,
                 validTo: datetime = None,
                 isFinal: bool = None, isExternalReference: bool = None,
                 serviceUrl: str = None,
                 structureUrl: str = None, maintainer=None,
                 items=None):

        super(AgencyScheme, self). \
            __init__(id_=id_, uri=uri, urn=urn,
                     annotations=annotations,
                     name=name, description=description,
                     version=version,
                     validFrom=validFrom,
                     validTo=validTo,
                     isFinal=isFinal,
                     isExternalReference=isExternalReference,
                     serviceUrl=serviceUrl,
                     structureUrl=structureUrl,
                     maintainer=maintainer,
                     items=items)

    def __eq__(self, other):
        if isinstance(other, AgencyScheme):
            return super(AgencyScheme, self).__eq__(other)

        return False


class Item(NameableArtefact):
    """The Item is an item of content in an Item Scheme. This may be a
    concept in a concept scheme, a code in a codelist etc.
    """

    _schemeType = None

    def __init__(self, id_: str = None, uri: str = None, urn: str = None,
                 annotations=None,
                 name: InternationalString = None,
                 description: InternationalString = None, scheme=None,
                 parent=None,
                 childs=None):

        if childs is None:
            childs = []
        if annotations is None:
            annotations = []
        super(Item, self).__init__(id_=id_, uri=uri, urn=urn,
                                   annotations=annotations, name=name,
                                   description=description)

        self.scheme = scheme

        self.parent = parent

        self._childs = []
        urn_list = []
        for c in childs:
            if c.urn not in urn_list:
                urn_list.append(c.urn)
                self.add_child(c)
            else:
                raise ValueError('Item cannot have two childs with same URN')

    def __eq__(self, other):
        if isinstance(other, Item):
            return super(Item, self).__eq__(other)

        return False

    def __str__(self):
        return f'<{self.__class__.__name__} - {self.id}>'

    def __unicode__(self):
        return f'<{self.__class__.__name__} - {self.id}>'

    def __repr__(self):
        return f'<{self.__class__.__name__} - {self.id}>'

    @property
    def scheme(self):
        """Reference to the ItemScheme"""
        return self._scheme

    @property
    def parent(self):
        """Reference to the parent Item"""
        return self._parent

    @property
    def childs(self):
        """Reference to the child Items"""
        return self._childs

    @scheme.setter
    def scheme(self, value):
        if value is None:
            self._scheme = value
        elif value.__class__.__name__ == self._schemeType:
            value.append(self)
            self._scheme = value
        else:
            raise TypeError(f"The scheme object has to be of the type "
                            f"{self._schemeType}")

    @parent.setter
    def parent(self, value):
        self._parent = value

    def add_child(self, value):
        """Adds a child to the Item"""
        self._childs.append(value)

    def _parse_XML(self, indent, head):
        head = f'{structureAbbr}:' + head

        prettyprint = indent != ''

        data = super(Item, self)._to_XML(prettyprint)
        indent = add_indent(indent)
        outfile = f'{indent}<{head}{data["Attributes"]}>'
        outfile += export_intern_data(data, indent)
        if self.parent is not None:
            indent_par = add_indent(indent)
            indent_ref = add_indent(indent_par)
            outfile += f'{indent_par}<{structureAbbr}:Parent>'
            if isinstance(self.parent, Item):
                text = self.parent.id
            else:
                text = self.parent
            outfile += f'{indent_ref}<Ref id="{text}"/>'
            outfile += f'{indent_par}</{structureAbbr}:Parent>'
        return outfile


class Code(Item):
    """ A language independent set of letters, numbers or symbols that
    represent a concept whose meaning is described in a natural language

        SchemeType:
            :obj:`Codelist`

    """

    _schemeType = "Codelist"

    def __init__(self, id_: str = None, uri: str = None, urn: str = None,
                 annotations=None,
                 name: InternationalString = None,
                 description: InternationalString = None,
                 scheme: ItemScheme = None, parent: Item = None, childs=None):
        if childs is None:
            childs = []
        if annotations is None:
            annotations = []
        super(Code, self).__init__(id_=id_, uri=uri, urn=urn,
                                   annotations=annotations,
                                   name=name, description=description,
                                   scheme=scheme, parent=parent, childs=childs)

    def __eq__(self, other):
        if isinstance(other, Code):
            return super(Code, self).__eq__(other)

        return False

    def _parse_XML(self, indent, head):
        outfile = super(Code, self)._parse_XML(indent, head)

        indent = add_indent(indent)
        outfile += f'{indent}</{structureAbbr}:{head}>'

        return outfile


class Agency(Item):
    """ Provides identity to all derived classes. It also provides
    annotations to derived classes because it is a subclass of
    AnnotableArtefact.

        SchemeType:
            :obj:`AgencyScheme`
    """
    _schemeType = "AgencyScheme"

    def __init__(self, id_: str = None, uri: str = None, urn: str = None,
                 annotations=None,
                 name: InternationalString = None,
                 description: InternationalString = None,
                 scheme: ItemScheme = None, contacts: List[Contact] = None):
        if annotations is None:
            annotations = []
        super(Agency, self).__init__(id_=id_, uri=uri, urn=urn,
                                     annotations=annotations,
                                     name=name, description=description,
                                     scheme=scheme)

        self._contacts = contacts

    def __eq__(self, other):
        if isinstance(other, Agency):
            return super(Agency, self).__eq__(
                other) and self._contacts == other._contacts

        return False

    @property
    def contacts(self):
        """list of associations to the Contact Information"""
        return self._contacts

    @contacts.setter
    def contacts(self, value):
        self._contacts = generic_setter(value, List[Contact])

    def _parse_XML(self, indent, head):
        outfile = super(Agency, self)._parse_XML(indent, head)

        indent = add_indent(indent)

        if self.contacts is not None:
            indent_child = add_indent(indent)
            for e in self.contacts:
                outfile += f'{indent_child}<{structureAbbr}:Contact>'
                outfile += e.to_xml(indent_child)
                outfile += f'{indent_child}</{structureAbbr}:Contact>'

        outfile += f'{indent}</{structureAbbr}:{head}>'

        return outfile


class Concept(Item):
    """ A concept is a unit of knowledge created by a unique combination of
    characteristics.

        SchemeType:
            :obj:`ConceptScheme`
    """

    _schemeType = "ConceptScheme"

    def __init__(self, id_: str = None, uri: str = None, urn: str = None,
                 annotations=None,
                 name: InternationalString = None,
                 description: InternationalString = None,
                 scheme: ItemScheme = None, parent: Item = None, childs=None,
                 core_representation: Representation = None):
        if childs is None:
            childs = []
        if annotations is None:
            annotations = []
        super(Concept, self).__init__(id_=id_, uri=uri, urn=urn,
                                      annotations=annotations,
                                      name=name, description=description,
                                      scheme=scheme, parent=parent,
                                      childs=childs)

        self.core_representation = core_representation
        self._ref = None  # Attribute for storing the references to codelists.

    def __eq__(self, other):
        if isinstance(other, Concept):
            return (super(Concept, self).__eq__(other) and
                    self._coreRepresentation == other._coreRepresentation)

        return False

    @property
    def core_representation(self):
        """Associates a Representation"""
        return self._coreRepresentation

    @core_representation.setter
    def core_representation(self, value):
        from .representation import Representation
        self._coreRepresentation = generic_setter(value, Representation)

    def _parse_XML(self, indent, head):
        outfile = super(Concept, self)._parse_XML(indent, head)

        indent = add_indent(indent)
        indent_child = add_indent(indent)
        if self.core_representation is not None:
            indent_enum = add_indent(indent_child)
            indent_ref = add_indent(indent_enum)
            outfile += f'{indent_child}<{structureAbbr}:CoreRepresentation>'
            if self.core_representation.codelist is not None:
                outfile += f'{indent_enum}<{structureAbbr}:Enumeration>'
                if isinstance(self.core_representation.codelist, str):
                    agencyID, id_, version = split_unique_id(
                        self.core_representation.codelist)

                    outfile += f'{indent_ref}<Ref package="codelist" ' \
                               f'agencyID="{agencyID}" ' \
                               f'id="{id_}" ' \
                               f'version="{version}" class="Codelist"/>'
                else:
                    agencyID = self.core_representation.codelist.agencyID
                    id_ = self.core_representation.codelist.id
                    version = self.core_representation.codelist.version

                    outfile += f'{indent_ref}<Ref package="codelist" ' \
                               f'agencyID="{agencyID}" ' \
                               f'id="{id_}" ' \
                               f'version="{version}" class="Codelist"/>'

                outfile += f'{indent_enum}</{structureAbbr}:Enumeration>'
            else:
                label_format = "TextFormat"
                format_attributes = ' '

                if self.core_representation.type_ is not None:
                    format_attributes = f' textType=' \
                                        f'"{self.core_representation.type_}"'

                if self.core_representation.facets is not None:
                    for e in self.core_representation.facets:
                        format_attributes += f' {e.facet_type}=' \
                                             f'"{e.facet_value}"'

                outfile += f'{indent_enum}<{structureAbbr}:' \
                           f'{label_format}{format_attributes}/>'
            outfile += f'{indent_child}</{structureAbbr}:CoreRepresentation>'

        outfile += f'{indent}</{structureAbbr}:{head}>'

        return outfile



from xml.parsers.expat import ExpatError

import xmltodict

from sdmxthon.parsers.data_read import create_dataset
from sdmxthon.parsers.metadata_read import create_metadata
from sdmxthon.utils.parsing_words import SERIES, OBS, STRSPE, GENERIC, \
    STRREF, STRUCTURE, STRID, namespaces, HEADER, DATASET, REF, AGENCY_ID, \
    ID, VERSION, DIM_OBS, ALL_DIM, STRUCTURES, STR_USAGE
from sdmxthon.utils.xml_base import validate_doc, \
    process_string_to_read

options = {'process_namespaces': True,
           'namespaces': namespaces,
           'dict_constructor': dict,
           'attr_prefix': ''}


def parse_sdmx(result):
    datasets = dict()

    if STRSPE in result:
        global_mode = STRSPE
    elif GENERIC in result:
        global_mode = GENERIC
    else:
        global_mode = STRUCTURE

    if global_mode == STRUCTURE:
        # Parsing Structure
        return create_metadata(result[STRUCTURE][STRUCTURES])
    else:
        message = result[global_mode]
        if isinstance(message[DATASET], list):
            structures = {}
            # Relationship between structures and structure id
            for structure in message[HEADER][STRUCTURE]:
                structures[structure[STRID]] = structure
            for single_dataset in message[DATASET]:
                str_ref = single_dataset[STRREF]
                if SERIES in single_dataset:
                    metadata = get_dataset_metadata(structures[str_ref],
                                                    str_ref,
                                                    mode=SERIES)
                else:
                    metadata = get_dataset_metadata(structures[str_ref],
                                                    str_ref,
                                                    mode=OBS)
                ds = create_dataset(single_dataset, metadata,
                                    global_mode)
                datasets[metadata[STRID]] = ds
        else:
            if SERIES in message[DATASET]:
                metadata = get_dataset_metadata(message[HEADER][STRUCTURE],
                                                message[DATASET][STRREF],
                                                mode=SERIES)
            elif OBS in message[DATASET]:
                metadata = get_dataset_metadata(message[HEADER][STRUCTURE],
                                                message[DATASET][STRREF],
                                                mode=OBS)
            else:
                if message[HEADER][STRUCTURE][DIM_OBS] == "AllDimensions":
                    mode = OBS
                else:
                    mode = SERIES
                metadata = get_dataset_metadata(message[HEADER][STRUCTURE],
                                                message[DATASET][STRREF],
                                                mode=mode)
            ds = create_dataset(message[DATASET], metadata, global_mode)
            datasets[metadata[STRID]] = ds

    return datasets


def read_xml(infile, mode=None, validate=True):
    infile = process_string_to_read(infile)

    if validate:
        validate_doc(infile)
    try:
        result = xmltodict.parse(infile, **options)
    except ExpatError as e:
        if e.offset > 10:  # UTF-8 BOM
            raise e
        result = xmltodict.parse(infile[3:], **options)

    del infile

    if mode is not None:
        if mode == "Data" and STRUCTURE in result:
            raise TypeError("Unable to parse metadata file as data file")
        elif mode == "Metadata" and (STRSPE in result or GENERIC in result):
            raise TypeError("Unable to parse data file as metadata file")
        elif mode not in ["Data", "Metadata"]:
            raise ValueError("Wrong mode")

    datasets = parse_sdmx(result)
    return datasets


def get_elements_from_structure(structure):
    if STRUCTURE in structure:
        agency_id = structure[STRUCTURE][REF][AGENCY_ID]
        id_ = structure[STRUCTURE][REF][ID]
        version = structure[STRUCTURE][REF][VERSION]
    else:
        agency_id = structure[STR_USAGE][REF][AGENCY_ID]
        id_ = structure[STR_USAGE][REF][ID]
        version = structure[STR_USAGE][REF][VERSION]
    return agency_id, id_, version


def get_dataset_metadata(structure, dataset_ref, mode):
    if mode == SERIES and structure[DIM_OBS] == ALL_DIM:
        raise Exception
    elif mode == OBS and structure[DIM_OBS] != ALL_DIM:
        raise Exception

    if dataset_ref == structure[STRID]:
        agency_id, id_, version = get_elements_from_structure(structure)
        return {DIM_OBS: structure[DIM_OBS],
                STRID: f"{agency_id}:{id_}({version})"}
    else:
        raise Exception



"""
    Utils has some handling functions for the model module
"""

import re
from datetime import datetime
from typing import List


#
# Convenience setters and getters
#

def set_date_from_string(value: str, format_: str = "%Y-%m-%dT%H:%M:%S"):
    """Generic function to format a string to datetime

    Args: value: The value to be validated.
    format_: A regex pattern to validate if the string has a specific format

    Returns:
        A datetime object

    Raises:
        ValueError: If the value violates the format constraint.
    """

    if value is None:
        return None
    if isinstance(value, datetime):
        return value
    if value[-1] == "Z":
        value = value[:-1]
    for fmt in (format_, "%Y-%m-%d", "%Y-%m-%dT%H:%M:%S",
                "%Y-%m-%dT%H:%M:%S.%f"):
        try:
            return datetime.strptime(value, fmt)
        except ValueError:
            pass

    raise ValueError(f"Wrong date string format. The format {format_} "
                     f"should be followed. {str(value)} passed")


def get_date_string(date: datetime, format_: str = "%Y-%m-%d"):
    """Generic function to get a string from a datetime object

    Args:
        date: The value to be validated.
        format_: A regex pattern to return the string with

    Returns:
        A string formatted

    Raises:
        ValueError: If the value violates any of the conditions.
    """
    if date is None:
        return ""

    return datetime.strftime(date, format_)


def string_setter(value: str, pattern: str = None,
                  enumeration: List[str] = None):
    """Generic function validating strings for setters

    Checks that the input is a string or integer.
    If it is a string, and a pattern is passed, checks that the pattern
    is respected.

    Args:
        value: The value to be validated.
        pattern: A regex pattern to be validated
        enumeration: A list with valid strings

    Returns:
        The validated string

    Raises:
        ValueError: If the value violates any of the conditions.
    """

    if isinstance(value, str):
        if pattern is not None:
            regex = re.compile(pattern, re.I)
            if regex.match(value):
                return value
            raise ValueError(f"Error setting the string. Pattern "
                             f"'{pattern}' not respected")
        elif enumeration is not None:
            if value in enumeration:
                return value
            raise ValueError(
                f"Error setting the string. Enumeration "
                f"{str(enumeration)} not respected")
        else:
            return value
    elif isinstance(value, int):
        return str(value)
    elif value is None:
        return None
    else:
        raise ValueError(f"Type should be a string, {type(value)} passed")


def date_setter(value: datetime):
    """Generic setter for datetime objects

    Args:
        value: The value to be validated.

    Raises:
        TypeError: If the value is not datetime or date
    """

    if isinstance(value, datetime) or value is None:
        return value
    if isinstance(value, str):
        return set_date_from_string(value)

    raise TypeError("Type should be datetime or date")


def bool_setter(value: bool):
    """Generic setter for bool objects

        Args:
            value: The value to be validated.

        Raises:
            TypeError: If the value is not bool
    """
    if isinstance(value, bool) or value is None:
        return value
    if value == "false":
        return False
    if value == "true":
        return True

    raise ValueError("Type should be bool")


def generic_setter(value, class_):
    """Generic setter for class objects

        Args:
            value: The value to be validated.
            class_: The class to validate with

        Raises:
            TypeError: If the value is not an instance of the class
    """
    if isinstance(value, class_) or value is None:
        return value

    raise TypeError(f"The value has to be an instance of the "
                    f"{class_.__name__} class. {type(value)} passed")


def int_setter(value: int):
    """Generic setter for integer objects

            Args:
                value: The value to be validated.

            Raises:
                TypeError: If the value is not an integer
        """

    if isinstance(value, int) or value is None:
        return value
    try:
        return int(value)
    except Exception:
        raise ValueError("Type should be int")


_base_ns = 'http://www.sdmx.org/resources/sdmxml/schemas/v2_1'
NS = {
    'com': f'{_base_ns}/common',
    'data': f'{_base_ns}/data/structurespecific',
    'str': f'{_base_ns}/structure',
    'mes': f'{_base_ns}/message',
    'gen': f'{_base_ns}/data/generic',
    'footer': f'{_base_ns}/message/footer',
    'xml': 'http://www.w3.org/XML/1998/namespace',
    'xsi': 'http://www.w3.org/2001/XMLSchema-instance',
}

ActionType = ['Delete, Replace, Append, Information']
UsageStatus = ['UsageStatus', 'mandatory conditional']
FacetValueType = ['string', 'bigInteger', 'integer', 'long', 'short',
                  'decimal', 'float', 'double', 'boolean', 'uri',
                  'count', 'inclusiveValueRange', 'alpha', 'alphaNumeric',
                  'numeric', 'exclusiveValueRange',
                  'incremental', 'observationalTimePeriod',
                  'standardTimePeriod', 'basicTimePeriod',
                  'gregorianTimePeriod', 'gregorianYear', 'gregorianMonth',
                  'gregorianYearMonth',
                  'gregorianDay', 'reportingTimePeriod', 'reportingYear',
                  'reportingSemester',
                  'reportingTrimester', 'reportingQuarter', 'reportingMonth',
                  'reportingWeek',
                  'reportingDay', 'dateTime', 'timesRange', 'month',
                  'monthDay', 'day', 'time', 'duration', 'keyValues',
                  'identifiableReference', 'dataSetReference', 'Xhtml']
ConstraintRoleType = ['Allowed', 'Actual']
FacetType = ['isSequence', 'minLength', 'maxLength', 'minValue', 'maxValue',
             'startValue', 'endValue', 'interval', 'timeInterval', 'decimals',
             'pattern', 'startTime', 'endTime']



from enum import Enum


class MessageTypeEnum(Enum):
    GenericDataSet = 'GenericDataSet'
    StructureDataSet = 'StructureDataSet'
    GenericTimeSeriesDataSet = 'GenericTimeSeriesDataSet'
    StructureTimeSeriesDataSet = 'StructureTimeSeriesDataSet'
    Metadata = 'Structures'



from zipfile import ZipFile

from sdmxthon.model.dataset import Dataset
from sdmxthon.model.message import Message
from sdmxthon.parsers.read import read_xml
from sdmxthon.utils.enums import MessageTypeEnum
from sdmxthon.utils.handlers import first_element_dict


def read_sdmx(sdmx_file, validate=True) -> Message:
    """
    Read SDMX performs the operation of reading a SDMX Data and SDMX
    metadata files. URLs could be used.

    :param sdmx_file: Path, URL or SDMX file as string
    :param validate: Validation of the XML file against the XSD (default: True)

    :return: A :obj:`Message <model.message.Message>` object
    """

    data = read_xml(sdmx_file, None, validate=validate)

    if isinstance(first_element_dict(data), Dataset):
        type_ = MessageTypeEnum.StructureDataSet
    else:
        type_ = MessageTypeEnum.Metadata
    return Message(type_, data)


def get_datasets(data, path_to_metadata, validate=True):
    """
    GetDatasets performs the operation of reading a SDMX Data and SDMX
    metadata files. URLs could be used.

    :param data: Path, URL or SDMX data file as string

    :param path_to_metadata: Path or URL to the SDMX metadata file

    :param validate: Validation of the XML file against the XSD (default: True)


    :return: A :obj:`Dataset <model.dataSet.DataSet>` object or a dict of \
    :obj:`Datasets <model.dataSet.DataSet>`
    """

    datasets = read_xml(data, mode="Data", validate=validate)

    metadata = read_xml(path_to_metadata,
                        mode="Metadata",
                        validate=validate)

    for v in datasets:
        if v in metadata['DataStructures']:
            datasets[v].structure = metadata['DataStructures'][v]
        elif v in metadata['Dataflows']:
            datasets[v].dataflow = metadata['Dataflows'][v]

    if len(datasets) == 1:
        return first_element_dict(datasets)

    return datasets


def get_pandas_df(data, validate=True):
    """
    GetPandasDF reads all observations in a SDMX file as Pandas Dataframe(s)

    :param data: Path, URL or SDMX data file as string

    :param validate: Validation of the XML file against the XSD (default: True)

    :return: A dict of `Pandas Dataframe \
    <https://pandas.pydata.org/docs/reference/api/pandas.DataFrame.html>`_
    """
    datasets = read_xml(data, "Data", validate=validate)
    return {ds: datasets[ds].data for ds in datasets}


def xml_to_csv(data, output_path=None, validate=True, **kwargs):
    """
    XML to CSV transforms a SDMX file into a CSV. Saves the file on disk or
    .zip of CSV. If the SDMX data file has only a Dataset and output_path is
    '', it returns a StringIO object. Kwargs are supported.

    :param data: Path, URL or SDMX file as string (Data file)
    :param output_path: Path to save the CSV (default: None)
    :param validate: Validation of the XML file against the XSD (default: True)
    :return: A StringIO object if output_path is ''
    """
    datasets = read_xml(data, mode="Data", validate=validate)

    if output_path is not None and '.zip' in output_path:
        with ZipFile(output_path, 'w') as zipObj:
            # Add multiple files to the zip
            for record in datasets:
                zipObj.writestr(record + '.csv',
                                data=datasets[record].to_csv(**kwargs))

    else:
        if len(datasets) > 1:
            raise ValueError('Cannot introduce several Datasets in a CSV. '
                             'Consider using .zip in output path')
        if len(datasets) == 1:
            if output_path is not None and '.zip' in output_path:
                filename = output_path.split('.')[0]
                output_path = filename + '.csv'
            # Getting first value
            dataset = first_element_dict(datasets)

            return dataset.to_csv(output_path, **kwargs)

        raise ValueError('No Datasets were parsed')



import io
import os
from io import BytesIO

import requests
import validators
from lxml import etree
from lxml.etree import DocumentInvalid

pathToSchema = 'schemas/SDMXMessage.xsd'


def process_string_to_read(infile: str):
    if isinstance(infile, str):
        if validators.url(infile):
            try:
                response = requests.get(infile)
                if response.status_code == 400:
                    raise requests.ConnectionError(
                        f'Invalid URL. Response from server: {response.text}')
                infile = io.TextIOWrapper(BytesIO(response.content),
                                          encoding='utf-8',
                                          errors="replace").read()
            except requests.ConnectionError:
                raise requests.ConnectionError('Invalid URL. '
                                               'No response from server')
        elif len(infile) > 10 and "<?" in infile[:10] and "xml" in infile[:10]:
            pass
        elif '/' in infile or '\\' in infile:
            try:
                infile = os.path.join(infile)
                with open(infile, "r", encoding='utf-8',
                          errors='replace') as f:
                    infile = f.read()
            except AttributeError:
                pass
        else:
            raise ValueError(f'Unable to parse {infile}')
    elif isinstance(infile, os.PathLike):
        try:
            infile = os.path.join(infile)
            with open(infile, "r", errors='replace',
                      encoding="utf-8") as f:
                infile = f.read()
        except AttributeError:
            pass
    elif isinstance(infile, BytesIO):
        infile = io.TextIOWrapper(infile,
                                  encoding='utf-8',
                                  errors="replace").read()

    if infile[0] != '<' and infile[3] == '<':
        infile = infile[3:]

    return infile


def validate_doc(infile):
    # Use the lxml ElementTree compatible parser so that, e.g.,
    #   we ignore comments.
    try:
        parser = etree.ETCompatXMLParser()
    except AttributeError:
        # fallback to xml.etree
        parser = etree.XMLParser()

    base_path = os.path.dirname(os.path.dirname(__file__))
    schema = os.path.join(base_path, pathToSchema)
    xmlschema_doc = etree.parse(schema)
    xmlschema = etree.XMLSchema(xmlschema_doc)

    if isinstance(infile, str):
        infile = BytesIO(bytes(infile, "UTF_8"))

    doc = etree.parse(infile, parser=parser)

    if not xmlschema.validate(doc):
        try:
            xmlschema.assertValid(doc)
        except DocumentInvalid as e:
            if len(e.args) == 1 and \
                    'xsi:type' in e.args[0] or \
                    'abstract' in e.args[0]:
                pass
            else:
                raise e


def cast(typ, value):
    if typ is None or value is None:
        return value
    return typ(value)
